# coding=utf-8
"""Tests that publish python plugin repositories."""
import random
from random import choice

from pulp_smash.pulp3.bindings import monitor_task
from pulp_smash.pulp3.utils import (
    get_content,
    get_versions,
    modify_repo
)
from urllib.parse import urljoin

from pulp_python.tests.functional.constants import (
    PYTHON_CONTENT_NAME,
    PULP_CONTENT_BASE_URL,
    PYTHON_SM_PROJECT_SPECIFIER,
    PYTHON_SM_FIXTURE_RELEASES,
    PYTHON_SM_FIXTURE_CHECKSUMS,
    PYTHON_EGG_FILENAME,
    PYTHON_WHEEL_FILENAME,
)
from pulp_python.tests.functional.utils import (
    cfg,
    TestCaseUsingBindings,
    TestHelpersMixin,
    ensure_simple,
)
from pulp_python.tests.functional.utils import set_up_module as setUpModule  # noqa:F401

from pulpcore.client.pulp_python.exceptions import ApiException


class PublishAnyRepoVersionTestCase(TestCaseUsingBindings, TestHelpersMixin):
    """Test whether a particular repository version can be published.

    This test targets the following issues:

    * `Pulp #3324 <https://pulp.plan.io/issues/3324>`_
    * `Pulp Smash #897 <https://github.com/pulp/pulp-smash/issues/897>`_
    """

    def test_all(self):
        """Test whether a particular repository version can be published.

        1. Create a repository with at least 2 repository versions.
        2. Create a publication by supplying the latest ``repository_version``.
        3. Assert that the publication ``repository_version`` attribute points
           to the latest repository version.
        4. Create a publication by supplying the non-latest ``repository_version``.
        5. Assert that the publication ``repository_version`` attribute points
           to the supplied repository version.
        6. Assert that an exception is raised when providing two different
           repository versions to be published at same time.
        """
        remote = self._create_remote()
        repo = self._create_repo_and_sync_with_remote(remote)
        # Step 1
        for python_content in get_content(repo.to_dict())[PYTHON_CONTENT_NAME]:
            modify_repo(cfg, repo.to_dict(), add_units=[python_content])

        version_hrefs = tuple(ver["pulp_href"] for ver in get_versions(repo.to_dict()))
        non_latest = choice(version_hrefs[:-1])

        repo = self.repo_api.read(repo.pulp_href)
        # Step 2, 3
        pub_latest = self._create_publication(repo)
        self.assertEqual(pub_latest.repository_version, version_hrefs[-1])
        # Step 4, 5
        pub_version = self._create_publication(repo, version=non_latest)
        self.assertEqual(pub_version.repository_version, non_latest)

        # Step 6
        with self.assertRaises(ApiException):
            body = {"repository": repo.pulp_href, "repository_version": non_latest}
            self.publications_api.create(body)


class PublishDifferentPolicyContent(TestCaseUsingBindings, TestHelpersMixin):
    """Test whether a 'on_demand', 'immediate', or 'streamed' synced repository can be published.

    This test targets the following issues:

    * `Pulp #7128 <https://pulp.plan.io/issues/7128>`_
    """

    def test_on_demand(self):
        """Test whether a particular repository version can be published.

        1. Create a repository
        2. Create a remote with on_demand sync policy
        3. Sync
        4. Publish repository
        """
        remote = self._create_remote(policy="on_demand")
        repo = self._create_repo_and_sync_with_remote(remote)
        pub = self._create_publication(repo)

        self.assertEqual(pub.repository_version, repo.latest_version_href)

    def test_immediate(self):
        """Test if immediate synced content can be published."""
        remote = self._create_remote(policy="immediate")
        repo = self._create_repo_and_sync_with_remote(remote)
        pub = self._create_publication(repo)

        self.assertEqual(pub.repository_version, repo.latest_version_href)

    def test_streamed(self):
        """Test if streamed synced content can be published."""
        remote = self._create_remote(policy="streamed")
        repo = self._create_repo_and_sync_with_remote(remote)
        pub = self._create_publication(repo)

        self.assertEqual(pub.repository_version, repo.latest_version_href)

    def test_mixed(self):
        """Test if repository with mixed synced content can be published."""
        # Sync on demand content
        remote = self._create_remote(policy="on_demand", includes=PYTHON_SM_PROJECT_SPECIFIER)
        repo = self._create_repo_and_sync_with_remote(remote)
        pub = self._create_publication(repo)
        # Sync immediate content
        remote2 = self._create_remote(policy="immediate")
        repo = self._sync_repo(repo, remote=remote2.pulp_href)
        pub2 = self._create_publication(repo)

        self.assertEqual(pub.repository_version, f"{repo.versions_href}1/")
        self.assertEqual(pub2.repository_version, repo.latest_version_href)


class ListPublications(TestCaseUsingBindings, TestHelpersMixin):
    """
    This tests that publications can be listed.

    This also tests that associated distributions show up when listing a publication.

    # TODO Make test RBAC specific when RBAC is implemented
    """

    def test_multiple_list(self):
        """Test listing multiple publications."""
        repo = self._create_repository()
        publication = self._create_publication(repo)
        publication2 = self._create_publication(repo)

        publications = self.publications_api.list()
        self.assertEqual(publications.results, [publication2, publication])

    def test_none_list(self):
        """Test listing publications when empty."""
        publications = self.publications_api.list()
        self.assertEqual(publications.count, 0)

    def test_publication_distribution(self):
        """Test that a publication properly lists it's connected distributions."""
        repo = self._create_repository()
        pub = self._create_publication(repo)
        distro = self._create_distribution_from_publication(pub)
        distro2 = self._create_distribution_from_publication(pub)

        publication = self.publications_api.read(pub.pulp_href)
        self.assertEqual(publication.distributions, [distro.pulp_href, distro2.pulp_href])


class DeletePublications(TestCaseUsingBindings, TestHelpersMixin):
    """
    These tests are dedicated to making sure that deleting works correctly.

    Deleted publications should be removed from associated distributions
    Deleted repositories should remove publications
    """

    def test_basic_delete(self):
        """Checks that deleting decreases publication count."""
        publications = self.publications_api.list()
        current_length = publications.count
        repo = self._create_repository()
        publication = self._create_publication(repo, cleanup=False)

        publications = self.publications_api.list()
        self.assertEqual(publications.count, 1 + current_length)

        self.publications_api.delete(publication.pulp_href)
        publications = self.publications_api.list()
        self.assertEqual(publications.count, current_length)

    def test_distro_removed_on_delete(self):
        """Test that distributions no longer point to a publication after deleted."""
        repo = self._create_repository()
        publication = self._create_publication(repo, cleanup=False)
        distro = self._create_distribution_from_publication(publication)

        self.assertEqual(distro.publication, publication.pulp_href)

        self.publications_api.delete(publication.pulp_href)
        distro = self.distributions_api.read(distro.pulp_href)

        self.assertEqual(distro.publication, None)

    def test_repository_delete(self):
        """Tests that deleting the repository of a publication deletes the publication."""
        repo = self._create_repository(cleanup=False)
        publication = self._create_publication(repo, cleanup=False)

        monitor_task(self.repo_api.delete(repo.pulp_href).task)
        with self.assertRaises(ApiException):
            self.publications_api.read(publication.pulp_href)


class PublishedCorrectContent(TestCaseUsingBindings, TestHelpersMixin):
    """
    These tests ensure that publications only serve content currently in the repository version.

    This test targets the following issues:

    * `Pulp #362 <https://github.com/pulp/pulp_python/issues/362>`_
    """

    @classmethod
    def setUpClass(cls):
        """Sets up the class variables"""
        super().setUpClass()
        cls.releases = [PYTHON_EGG_FILENAME, PYTHON_WHEEL_FILENAME]

    def test_all_content_published(self):
        """Publishes SM Project and ensures correctness of simple api."""
        remote = self._create_remote(includes=PYTHON_SM_PROJECT_SPECIFIER)
        repo = self._create_repo_and_sync_with_remote(remote)
        pub = self._create_publication(repo)
        distro = self._create_distribution_from_publication(pub)

        url = urljoin(PULP_CONTENT_BASE_URL, f"{distro.base_path}/simple/")
        proper, msgs = ensure_simple(url, PYTHON_SM_FIXTURE_RELEASES,
                                     sha_digests=PYTHON_SM_FIXTURE_CHECKSUMS)
        self.assertTrue(proper, msg=msgs)

    def test_removed_content_not_published(self):
        """Ensure content removed from a repository doesn't get published again."""
        remote = self._create_remote()
        repo = self._create_repo_and_sync_with_remote(remote)
        pub = self._create_publication(repo)
        distro = self._create_distribution_from_publication(pub)

        url = urljoin(PULP_CONTENT_BASE_URL, f"{distro.base_path}/simple/")
        proper, msgs = ensure_simple(url, {"shelf-reader": self.releases})
        self.assertTrue(proper, msg=msgs)

        removed_content = random.choice(get_content(repo.to_dict())[PYTHON_CONTENT_NAME])
        modify_repo(cfg, repo.to_dict(), add_units=[removed_content])
        repo = self.repo_api.read(repo.pulp_href)
        pub = self._create_publication(repo)
        distro = self._create_distribution_from_publication(pub)

        if removed_content["filename"] == PYTHON_WHEEL_FILENAME:
            remaining_release = [PYTHON_EGG_FILENAME]
        else:
            remaining_release = [PYTHON_WHEEL_FILENAME]
        url = urljoin(PULP_CONTENT_BASE_URL, f"{distro.base_path}/simple/")
        proper, msgs = ensure_simple(url, {"shelf-reader": remaining_release})

        self.assertTrue(proper, msg=msgs)

    def test_new_content_is_published(self):
        """Ensures added content is published with a new publication."""
        remote = self._create_remote(package_types=["sdist"])
        repo = self._create_repo_and_sync_with_remote(remote)
        pub = self._create_publication(repo)
        distro = self._create_distribution_from_publication(pub)

        url = urljoin(PULP_CONTENT_BASE_URL, f"{distro.base_path}/simple/")
        proper, m = ensure_simple(url, {"shelf-reader": [PYTHON_EGG_FILENAME]})
        self.assertTrue(proper, msg=m)

        remote = self._create_remote()
        repo = self._sync_repo(repo, remote=remote.pulp_href)
        pub = self._create_publication(repo)
        distro = self._create_distribution_from_publication(pub)

        url = urljoin(PULP_CONTENT_BASE_URL, f"{distro.base_path}/simple/")
        proper, msgs = ensure_simple(url, {"shelf-reader": self.releases})
        self.assertTrue(proper, msg=msgs)



# coding=utf-8
"""Utilities for tests for the python plugin."""
from functools import partial
from unittest import SkipTest, TestCase
from tempfile import NamedTemporaryFile
from urllib.parse import urljoin
from lxml import html

from pulp_smash import config, selectors
from pulp_smash.utils import http_get
from pulp_smash.pulp3.bindings import monitor_task
from pulp_smash.pulp3.utils import (
    gen_distribution,
    gen_remote,
    gen_repo,
    get_content,
    require_pulp_3,
    require_pulp_plugins,
)

from pulp_python.tests.functional.constants import (
    PYTHON_CONTENT_NAME,
    PYTHON_FIXTURE_URL,
    PYTHON_URL,
    PYTHON_EGG_FILENAME,
    PYTHON_XS_PROJECT_SPECIFIER,
)

from pulpcore.client.pulpcore import (
    ApiClient as CoreApiClient,
    ArtifactsApi,
    TasksApi,
)
from pulpcore.client.pulp_python import ApiClient as PythonApiClient
from pulpcore.client.pulp_python import (
    DistributionsPypiApi,
    RepositoriesPythonApi,
    ContentPackagesApi,
    PublicationsPypiApi,
    PythonPythonPublication,
    RemotesPythonApi,
    RepositorySyncURL,
)

cfg = config.get_config()
configuration = cfg.get_bindings_config()


def set_up_module():
    """Skip tests Pulp 3 isn't under test or if pulp_python isn't installed."""
    require_pulp_3(SkipTest)
    require_pulp_plugins({"python"}, SkipTest)


def gen_python_client():
    """Return an OBJECT for python client."""
    return PythonApiClient(configuration)


def gen_python_remote(url=PYTHON_FIXTURE_URL, includes=None, **kwargs):
    """Return a semi-random dict for use in creating a python Remote.

    :param url: The URL of an external content source.
    :param includes: An iterable of dicts containing project specifier dicts.
    :param **kwargs: Specified parameters for the Remote
    """
    remote = gen_remote(url)
    if includes is None:
        includes = PYTHON_XS_PROJECT_SPECIFIER

    # Remote also supports "excludes" and "prereleases".
    python_extra_fields = {
        "includes": includes,
        **kwargs,
    }
    remote.update(python_extra_fields)
    return remote


def get_python_content_paths(repo, version_href=None):
    """Return the relative path of content units present in a python repository.

    :param repo: A dict of information about the repository.
    :param version_href: The repository version to read.
    :returns: A dict of lists with the paths of units present in a given repository.
        Paths are given as pairs with the remote and the local version for different content types.
    """
    return {
        PYTHON_CONTENT_NAME: [
            (content_unit["filename"], content_unit["filename"])
            for content_unit in get_content(repo, version_href)[PYTHON_CONTENT_NAME]
        ],
    }


def gen_python_content_attrs(artifact, filename=PYTHON_EGG_FILENAME):
    """Generate a dict with content unit attributes.

    :param artifact: A dict of info about the artifact.
    :param filename: the name of the artifact being uploaded
    :returns: A semi-random dict for use in creating a content unit.
    """
    return {
        "artifact": artifact["pulp_href"],
        "relative_path": filename,
    }


core_client = CoreApiClient(configuration)
tasks = TasksApi(core_client)
py_client = gen_python_client()
repo_api = RepositoriesPythonApi(py_client)
remote_api = RemotesPythonApi(py_client)
pub_api = PublicationsPypiApi(py_client)
distributions_api = DistributionsPypiApi(py_client)
content_api = ContentPackagesApi(py_client)


def populate_pulp(url=PYTHON_FIXTURE_URL):
    """Add python contents to Pulp.

    :param pulp_smash.config.PulpSmashConfig: Information about a Pulp application.
    :param url: The python repository URL. Defaults to
        :data:`pulp_smash.constants.PYTHON_FIXTURE_URL`
    :returns: A list of dicts, where each dict describes one python content in Pulp.
    """
    remote = None
    repo = None
    try:
        remote = remote_api.create(gen_python_remote(url))
        repo = repo_api.create(gen_repo())

        repository_sync_data = RepositorySyncURL(remote=remote.pulp_href)
        sync_response = repo_api.sync(repo.pulp_href, repository_sync_data)
        monitor_task(sync_response.task)
    finally:
        if remote:
            remote_api.delete(remote.pulp_href)
        if repo:
            repo_api.delete(repo.pulp_href)
    return content_api.list().to_dict()["results"]


def publish(repo, version_href=None):
    """Publish a repository.
    :param repo: A dict of information about the repository.
    :param version_href: A href for the repo version to be published.
    :returns: A publication. A dict of information about the just created
        publication.
    """
    if version_href:
        publish_data = PythonPythonPublication(repository_href=version_href)
    else:
        publish_data = PythonPythonPublication(repository=repo["pulp_href"])

    publish_response = pub_api.create(publish_data)
    created_resources = monitor_task(publish_response.task).created_resources
    return pub_api.read(created_resources[0]).to_dict()


skip_if = partial(selectors.skip_if, exc=SkipTest)  # pylint:disable=invalid-name
"""The ``@skip_if`` decorator, customized for unittest.

:func:`pulp_smash.selectors.skip_if` is test runner agnostic. This function is
identical, except that ``exc`` has been set to ``unittest.SkipTest``.
"""


def gen_artifact(url=PYTHON_URL):
    """Creates an artifact."""
    with NamedTemporaryFile() as temp_file:
        temp_file.write(http_get(url))
        temp_file.flush()
        artifact = ArtifactsApi(core_client).create(file=temp_file.name)
        return artifact.to_dict()


class TestCaseUsingBindings(TestCase):
    """A parent TestCase that instantiates the various bindings used throughout tests."""

    @classmethod
    def setUpClass(cls):
        """Create class-wide variables."""
        cls.client = py_client
        cls.core_client = core_client
        cls.repo_api = repo_api
        cls.remote_api = remote_api
        cls.publications_api = pub_api
        cls.distributions_api = distributions_api
        cls.content_api = content_api


class TestHelpersMixin:
    """A common place for sync helper functions."""

    def _create_repository(self, cleanup=True, **kwargs):
        """Create `PythonRepository`"""
        repo = self.repo_api.create(gen_repo(**kwargs))
        if cleanup:
            self.addCleanup(self.repo_api.delete, repo.pulp_href)
        return repo

    def _create_remote(self, cleanup=True, **kwargs):
        """Create `PythonRemote`."""
        remote = self.remote_api.create(gen_python_remote(**kwargs))
        if cleanup:
            self.addCleanup(self.remote_api.delete, remote.pulp_href)
        return remote

    def _create_repo_and_sync_with_remote(self, remote, **kwargs):
        """
        Create a repository and then sync with the provided `remote`.

        Args:
            remote: The remote to be sync with

        Returns:
            repository: The created repository object to be asserted to.
        """
        # Create the repository.
        repo = self.repo_api.create(gen_repo(remote=remote.pulp_href))
        self.addCleanup(self.repo_api.delete, repo.pulp_href)
        return self._sync_repo(repo, remote=remote.pulp_href, **kwargs)

    def _create_repo_with_attached_remote_and_sync(self, remote, **kwargs):
        """
        Create a repository with the remote attached, and then sync without specifying the `remote`.

        Args:
            remote: The remote to attach to the repository

        Returns:
            repository: The created repository object to be asserted to.
        """
        # Create the repository.
        repo = self.repo_api.create(gen_repo(remote=remote.pulp_href))
        self.addCleanup(self.repo_api.delete, repo.pulp_href)
        return self._sync_repo(repo, **kwargs)

    def _sync_repo(self, repo, **kwargs):
        """
        Sync the repo with optional `kwarg` parameters passed on to the sync method.

        Args:
            repo: The repository to sync

        Returns:
            repository: The updated repository after the sync is complete
        """
        repository_sync_data = RepositorySyncURL(**kwargs)
        sync_response = self.repo_api.sync(repo.pulp_href, repository_sync_data)
        monitor_task(sync_response.task)
        repo = self.repo_api.read(repo.pulp_href)
        return repo

    def _create_distribution(self, cleanup=True, **kwargs):
        """Create a `PythonDistribution`."""
        body = gen_distribution(**kwargs)
        distribution_create = self.distributions_api.create(body)
        created_resources = monitor_task(distribution_create.task).created_resources
        distribution = self.distributions_api.read(created_resources[0])
        if cleanup:
            self.addCleanup(self.distributions_api.delete, distribution.pulp_href)
        return distribution

    def _create_distribution_from_repo(self, repo, cleanup=True):
        """
        Create a `PythonDistribution` serving the `repo`.

        Args:
            repo: The repository to serve with the `PythonDistribution`
            cleanup: Whether the distribution should be cleaned up

        Returns:
            The created `PythonDistribution`.
        """
        # Create a distribution.
        return self._create_distribution(cleanup, repository=repo.pulp_href)

    def _create_empty_repo_and_distribution(self, cleanup=True, **repo_kwargs):
        """
        Creates an empty `PythonRepository` and an `PythonDistribution` serving that repository.

        Args:
            cleanup: Whether the repository and distribution should be cleaned up

        Returns:
            Tuple of the created `PythonRepository`, `PythonDistribution`
        """
        repo = self.repo_api.create(gen_repo(**repo_kwargs))
        if cleanup:
            self.addCleanup(self.repo_api.delete, repo.pulp_href)
        return repo, self._create_distribution_from_repo(repo, cleanup=cleanup)

    def _create_publication(self, repository, version=None, cleanup=True):
        """
        Creates a `PythonPublication` from the `PythonRepository`.

        Args:
            repository: The repository to use for the publication
            version: Optional - The version to repo_version to use, either an int or href
            cleanup: whether to cleanup the publication

        Returns:
            The created `PythonPublication`
        """
        if version is not None:
            try:
                rv_href = f"{repository.versions_href}{int(version)}/"
            except ValueError:
                rv_href = version
            publish_data = PythonPythonPublication(repository_version=rv_href)
        else:
            publish_data = PythonPythonPublication(repository=repository.pulp_href)
        publish_response = self.publications_api.create(publish_data)
        pub = self.publications_api.read(monitor_task(publish_response.task).created_resources[0])
        if cleanup:
            self.addCleanup(self.publications_api.delete, pub.pulp_href)
        return pub

    def _create_distribution_from_publication(self, pub, cleanup=True):
        """
        Create an `PythonDistribution` serving the `pub`.

        Args:
            pub: The publication to serve with the `PythonDistribution`
            repo: Optional repo to create the publication from if pub is not set
            version: Optional version to create the publication from if pub is not set
            cleanup: Whether the distribution should be cleaned up

        Returns:
            The created `PythonDistribution`.
        """
        return self._create_distribution(cleanup, publication=pub.pulp_href)


def ensure_simple(simple_url, packages, sha_digests=None):
    """
    Tests that the simple api at `url` matches the packages supplied.
    `packages`: dictionary of form {package_name: [release_filenames]}
    First check `/simple/index.html` has each package name, no more, no less
    Second check `/simple/package_name/index.html` for each package exists
    Third check each package's index has all their releases listed, no more, no less
    Returns tuple (`proper`: bool, `error_msg`: str)
    *Technically, if there was a bug, other packages' indexes could be posted, but not present
    in the simple index and thus be accessible from the distribution, but if one can't see it
    how would one know that it's there?*
    """
    def explore_links(page_url, page_name, links_found, msgs):
        legit_found_links = []
        page = html.fromstring(http_get(page_url))
        page_links = page.xpath("/html/body/a")
        for link in page_links:
            if link.text in links_found:
                if links_found[link.text]:
                    msgs += f"\nDuplicate {page_name} name {link.text}"
                links_found[link.text] = True
                if link.get("href"):
                    legit_found_links.append(link.get("href"))
                else:
                    msgs += f"\nFound {page_name} link without href {link.text}"
            else:
                msgs += f"\nFound extra {page_name} link {link.text}"
        return legit_found_links

    packages_found = {name: False for name in packages.keys()}
    releases_found = {name: False for releases in packages.values() for name in releases}
    msgs = ""
    found_release_links = explore_links(simple_url, "simple", packages_found, msgs)
    dl_links = []
    for rel_link in found_release_links:
        dl_links += explore_links(urljoin(simple_url, rel_link), "release", releases_found, msgs)
    for dl_link in dl_links:
        package_link, _, sha = dl_link.partition("#sha256=")
        if len(sha) != 64:
            msgs += f"\nRelease download link has bad sha256 {dl_link}"
        if sha_digests:
            package = package_link.split("/")[-1]
            if sha_digests[package] != sha:
                msgs += f"\nRelease has bad sha256 attached to it {package}"
    msgs += "".join(map(lambda x: f"\nSimple link not found for {x}",
                        [name for name, val in packages_found.items() if not val]))
    msgs += "".join(map(lambda x: f"\nReleases link not found for {x}",
                        [name for name, val in releases_found.items() if not val]))
    return len(msgs) == 0, msgs



# Generated by Django 2.2.20 on 2021-04-26 16:28

from django.db import migrations, models, transaction


def add_sha256_to_current_models(apps, schema_editor):
    """Adds the sha256 to current PythonPackageContent models."""
    PythonPackageContent = apps.get_model('python', 'PythonPackageContent')
    RemoteArtifact = apps.get_model('core', 'RemoteArtifact')
    package_bulk = []
    for python_package in PythonPackageContent.objects.only("pk", "sha256").iterator():
        content_artifact = python_package.contentartifact_set.first()
        if content_artifact.artifact:
            artifact = content_artifact.artifact
        else:
            artifact = RemoteArtifact.objects.filter(content_artifact=content_artifact).first()
        python_package.sha256 = artifact.sha256
        package_bulk.append(python_package)
        if len(package_bulk) == 100000:
            with transaction.atomic():
                PythonPackageContent.objects.bulk_update(package_bulk, ["sha256",])
                package_bulk = []
    with transaction.atomic():
        PythonPackageContent.objects.bulk_update(package_bulk, ["sha256",])


class Migration(migrations.Migration):

    dependencies = [
        ('python', '0004_DATA_swap_distribution_model'),
    ]

    operations = [
        migrations.AddField(
            model_name='pythonpackagecontent',
            name='sha256',
            field=models.CharField(max_length=64, default=''),
            preserve_default=False,
        ),
        migrations.RunPython(add_sha256_to_current_models, migrations.RunPython.noop)
    ]



import pkginfo
import shutil
import tempfile
import json
from collections import defaultdict
from django.core.files.storage import default_storage as storage
from django.conf import settings
from jinja2 import Template
from packaging.utils import canonicalize_name
from packaging.version import parse


PYPI_LAST_SERIAL = "X-PYPI-LAST-SERIAL"
"""TODO This serial constant is temporary until Python repositories implements serials"""
PYPI_SERIAL_CONSTANT = 1000000000

simple_index_template = """<!DOCTYPE html>
<html>
  <head>
    <title>Simple Index</title>
    <meta name="api-version" value="2" />
  </head>
  <body>
    {% for name, canonical_name in projects %}
    <a href="{{ canonical_name }}/">{{ name }}</a><br/>
    {% endfor %}
  </body>
</html>
"""

simple_detail_template = """<!DOCTYPE html>
<html>
<head>
  <title>Links for {{ project_name }}</title>
  <meta name="api-version" value="2" />
</head>
<body>
    <h1>Links for {{ project_name }}</h1>
    {% for name, path, sha256 in project_packages %}
    <a href="{{ path }}#sha256={{ sha256 }}" rel="internal">{{ name }}</a><br/>
    {% endfor %}
</body>
</html>
"""

DIST_EXTENSIONS = {
    ".whl": "bdist_wheel",
    ".exe": "bdist_wininst",
    ".egg": "bdist_egg",
    ".tar.bz2": "sdist",
    ".tar.gz": "sdist",
    ".zip": "sdist",
}

DIST_TYPES = {
    "bdist_wheel": pkginfo.Wheel,
    "bdist_wininst": pkginfo.Distribution,
    "bdist_egg": pkginfo.BDist,
    "sdist": pkginfo.SDist,
}


def parse_project_metadata(project):
    """
    Create a dictionary of python project metadata.

    Args:
        project (dict): Metadata relevant to the entire Python project

    Returns:
        dictionary: of python project metadata

    """
    package = {}
    package['name'] = project.get('name') or ""
    package['metadata_version'] = project.get('metadata_version') or ""
    package['summary'] = project.get('summary') or ""
    package['description'] = project.get('description') or ""
    package['keywords'] = project.get('keywords') or ""
    package['home_page'] = project.get('home_page') or ""
    package['download_url'] = project.get('download_url') or ""
    package['author'] = project.get('author') or ""
    package['author_email'] = project.get('author_email') or ""
    package['maintainer'] = project.get('maintainer') or ""
    package['maintainer_email'] = project.get('maintainer_email') or ""
    package['license'] = project.get('license') or ""
    package['project_url'] = project.get('project_url') or ""
    package['platform'] = project.get('platform') or ""
    package['supported_platform'] = project.get('supported_platform') or ""
    package['requires_dist'] = json.dumps(project.get('requires_dist', []))
    package['provides_dist'] = json.dumps(project.get('provides_dist', []))
    package['obsoletes_dist'] = json.dumps(project.get('obsoletes_dist', []))
    package['requires_external'] = json.dumps(project.get('requires_external', []))
    package['classifiers'] = json.dumps(project.get('classifiers', []))
    package['project_urls'] = json.dumps(project.get('project_urls', {}))
    package['description_content_type'] = project.get('description_content_type') or ""

    return package


def parse_metadata(project, version, distribution):
    """
    Extract metadata from a distribution.

    Create a dictionary of metadata needed to create a PythonContentUnit from
    the project, version, and distribution metadata.

    Args:
        project (dict): Metadata relevant to the entire Python project
        version (string): Version of distribution
        distribution (dict): Metadata of a single Python distribution

    Returns:
        dictionary: of useful python metadata

    """
    package = {}

    package['filename'] = distribution.get('filename') or ""
    package['packagetype'] = distribution.get('packagetype') or ""
    package['version'] = version
    package['url'] = distribution.get('url') or ""
    package['sha256'] = distribution.get('digests', {}).get('sha256') or ""
    package['python_version'] = distribution.get('python_version') or ""
    package['requires_python'] = distribution.get('requires_python') or ""

    package.update(parse_project_metadata(project))

    return package


def get_project_metadata_from_artifact(filename, artifact):
    """
    Gets the metadata of a Python Package.

    Raises ValueError if filename has an unsupported extension
    """
    extensions = list(DIST_EXTENSIONS.keys())
    # Iterate through extensions since splitext does not support things like .tar.gz
    # If no supported extension is found, ValueError is raised here
    pkg_type_index = [filename.endswith(ext) for ext in extensions].index(True)
    packagetype = DIST_EXTENSIONS[extensions[pkg_type_index]]
    # Copy file to a temp directory under the user provided filename, we do this
    # because pkginfo validates that the filename has a valid extension before
    # reading it
    with tempfile.NamedTemporaryFile('wb', dir=".", suffix=filename) as temp_file:
        artifact_file = storage.open(artifact.file.name)
        shutil.copyfileobj(artifact_file, temp_file)
        temp_file.flush()
        metadata = DIST_TYPES[packagetype](temp_file.name)
        metadata.packagetype = packagetype
        return metadata


def python_content_to_json(base_path, content_query, version=None):
    """
    Converts a QuerySet of PythonPackageContent into the PyPi JSON format
    https://www.python.org/dev/peps/pep-0566/
    JSON metadata has:
        info: Dict
        last_serial: int
        releases: Dict
        urls: Dict

    Returns None if version is specified but not found within content_query
    """
    full_metadata = {"last_serial": 0}  # For now the serial field isn't supported by Pulp
    latest_content = latest_content_version(content_query, version)
    if not latest_content:
        return None
    full_metadata.update({"info": python_content_to_info(latest_content[0])})
    full_metadata.update({"releases": python_content_to_releases(content_query, base_path)})
    full_metadata.update({"urls": python_content_to_urls(latest_content, base_path)})
    return full_metadata


def latest_content_version(content_query, version):
    """
    Walks through the content QuerySet and finds the instances that is the latest version.
    If 'version' is specified, the function instead tries to find content instances
    with that version and will return an empty list if nothing is found
    """
    latest_version = version
    latest_content = []
    for content in content_query:
        if version and parse(version) == parse(content.version):
            latest_content.append(content)
        elif not latest_version or parse(content.version) > parse(latest_version):
            latest_content = [content]
            latest_version = content.version
        elif parse(content.version) == parse(latest_version):
            latest_content.append(content)

    return latest_content


def json_to_dict(data):
    """
    Converts a JSON string into a Python dictionary.

    Args:
        data (string): JSON string

    Returns:
        dictionary: of JSON string

    """
    if isinstance(data, dict):
        return data

    return json.loads(data)


def python_content_to_info(content):
    """
    Takes in a PythonPackageContent instance and returns a dictionary of the Info fields
    """
    return {
        "name": content.name,
        "version": content.version,
        "summary": content.summary or "",
        "keywords": content.keywords or "",
        "description": content.description or "",
        "description_content_type": content.description_content_type or "",
        "bugtrack_url": None,  # These two are basically never used
        "docs_url": None,
        "downloads": {"last_day": -1, "last_month": -1, "last_week": -1},
        "download_url": content.download_url or "",
        "home_page": content.home_page or "",
        "author": content.author or "",
        "author_email": content.author_email or "",
        "maintainer": content.maintainer or "",
        "maintainer_email": content.maintainer_email or "",
        "license": content.license or "",
        "requires_python": content.requires_python or None,
        "package_url": content.project_url or "",  # These two are usually identical
        "project_url": content.project_url or "",  # They also usually point to PyPI
        "release_url": f"{content.project_url}{content.version}/" if content.project_url else "",
        "project_urls": json_to_dict(content.project_urls) or None,
        "platform": content.platform or "",
        "requires_dist": json_to_dict(content.requires_dist) or None,
        "classifiers": json_to_dict(content.classifiers) or None,
        "yanked": False,  # These are no longer used on PyPI, but are still present
        "yanked_reason": None,
    }


def python_content_to_releases(content_query, base_path):
    """
    Takes a QuerySet of PythonPackageContent and returns a dictionary of releases
    with each key being a version and value being a list of content for that version of the package
    """
    releases = defaultdict(lambda: [])
    for content in content_query:
        releases[content.version].append(python_content_to_download_info(content, base_path))
    return releases


def python_content_to_urls(contents, base_path):
    """
    Takes the latest content in contents and returns a list of download information
    """
    return [python_content_to_download_info(content, base_path) for content in contents]


def python_content_to_download_info(content, base_path):
    """
    Takes in a PythonPackageContent and base path of the distribution to create a dictionary of
    download information for that content. This dictionary is used by Releases and Urls.
    """
    def find_artifact():
        _art = content_artifact.artifact
        if not _art:
            from pulpcore.plugin import models
            _art = models.RemoteArtifact.objects.filter(content_artifact=content_artifact).first()
        return _art

    content_artifact = content.contentartifact_set.first()
    artifact = find_artifact()
    origin = settings.CONTENT_ORIGIN.strip("/")
    prefix = settings.CONTENT_PATH_PREFIX.strip("/")
    base_path = base_path.strip("/")
    url = "/".join((origin, prefix, base_path, content.filename))
    return {
        "comment_text": "",
        "digests": {"md5": artifact.md5, "sha256": artifact.sha256},
        "downloads": -1,
        "filename": content.filename,
        "has_sig": False,
        "md5_digest": artifact.md5,
        "packagetype": content.packagetype,
        "python_version": content.python_version,
        "requires_python": content.requires_python or None,
        "size": artifact.size,
        "upload_time": str(artifact.pulp_created),
        "upload_time_iso_8601": str(artifact.pulp_created.isoformat()),
        "url": url,
        "yanked": False,
        "yanked_reason": None
    }


def write_simple_index(project_names, streamed=False):
    """Writes the simple index."""
    simple = Template(simple_index_template)
    context = {"projects": ((x, canonicalize_name(x)) for x in project_names)}
    return simple.stream(**context) if streamed else simple.render(**context)


def write_simple_detail(project_name, project_packages, streamed=False):
    """Writes the simple detail page of a package."""
    detail = Template(simple_detail_template)
    context = {"project_name": project_name, "project_packages": project_packages}
    return detail.stream(**context) if streamed else detail.render(**context)



from gettext import gettext as _
import logging
import os

from django.core.files import File
from packaging.utils import canonicalize_name

from pulpcore.plugin import models

from pulp_python.app import models as python_models
from pulp_python.app.utils import write_simple_index, write_simple_detail


log = logging.getLogger(__name__)


def publish(repository_version_pk):
    """
    Create a Publication based on a RepositoryVersion.

    Args:
        repository_version_pk (str): Create a Publication from this RepositoryVersion.

    """
    repository_version = models.RepositoryVersion.objects.get(pk=repository_version_pk)

    log.info(_('Publishing: repository={repo}, version={version}').format(
        repo=repository_version.repository.name,
        version=repository_version.number,
    ))

    with python_models.PythonPublication.create(repository_version, pass_through=True) as pub:
        write_simple_api(pub)

    log.info(_('Publication: {pk} created').format(pk=pub.pk))
    return pub


def write_simple_api(publication):
    """
    Write metadata for the simple API.

    Writes metadata mimicking the simple api of PyPI for all python packages
    in the repository version.

    https://wiki.python.org/moin/PyPISimple

    Args:
        publication (pulpcore.plugin.models.Publication): A publication to generate metadata for

    """
    simple_dir = 'simple/'
    os.mkdir(simple_dir)
    project_names = (
        python_models.PythonPackageContent.objects.filter(
            pk__in=publication.repository_version.content
        )
        .order_by('name')
        .values_list('name', flat=True)
        .distinct()
    )

    # write the root index, which lists all of the projects for which there is a package available
    index_path = '{simple_dir}index.html'.format(simple_dir=simple_dir)
    with open(index_path, 'w') as index:
        index.write(write_simple_index(project_names))

    index_metadata = models.PublishedMetadata.create_from_file(
        relative_path=index_path,
        publication=publication,
        file=File(open(index_path, 'rb'))
    )
    index_metadata.save()

    if len(project_names) == 0:
        return

    packages = python_models.PythonPackageContent.objects.filter(
        pk__in=publication.repository_version.content
    )
    releases = packages.order_by("name").values("name", "filename", "sha256")

    ind = 0
    current_name = project_names[ind]
    package_releases = []
    for release in releases.iterator():
        if release['name'] != current_name:
            write_project_page(
                name=canonicalize_name(current_name),
                simple_dir=simple_dir,
                package_releases=package_releases,
                publication=publication
            )
            package_releases = []
            ind += 1
            current_name = project_names[ind]
        relative_path = release['filename']
        path = f"../../{relative_path}"
        checksum = release['sha256']
        package_releases.append((relative_path, path, checksum))
    # Write the final project's page
    write_project_page(
        name=canonicalize_name(current_name),
        simple_dir=simple_dir,
        package_releases=package_releases,
        publication=publication
    )


def write_project_page(name, simple_dir, package_releases, publication):
    """Writes a project's simple page."""
    project_dir = f'{simple_dir}{name}/'
    os.mkdir(project_dir)
    metadata_relative_path = f'{project_dir}index.html'

    with open(metadata_relative_path, 'w') as simple_metadata:
        simple_metadata.write(write_simple_detail(name, package_releases))

    project_metadata = models.PublishedMetadata.create_from_file(
        relative_path=metadata_relative_path,
        publication=publication,
        file=File(open(metadata_relative_path, 'rb'))
    )
    project_metadata.save()  # change to bulk create when multi-table supported



# Generated by Django 2.2.23 on 2021-06-10 01:57

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('python', '0007_pythonpackagecontent_mv-2-1'),
    ]

    operations = [
        migrations.AlterField(
            model_name='pythonpackagecontent',
            name='filename',
            field=models.TextField(db_index=True),
        ),
        migrations.AlterField(
            model_name='pythonpackagecontent',
            name='sha256',
            field=models.CharField(db_index=True, max_length=64, unique=True),
        ),
        migrations.SeparateDatabaseAndState(
            state_operations=[
                migrations.AlterUniqueTogether(
                    name='pythonpackagecontent',
                    unique_together={('sha256',)},
                ),
            ],
        ),
    ]



import time

from datetime import datetime, timezone
from django.db import transaction
from django.contrib.sessions.models import Session
from pulpcore.plugin.models import Artifact, CreatedResource, ContentArtifact

from pulp_python.app.models import PythonPackageContent, PythonRepository
from pulp_python.app.utils import get_project_metadata_from_artifact, parse_project_metadata


def upload(artifact_sha256, filename, repository_pk=None):
    """
    Uploads a Python Package to Pulp

    Args:
        artifact_sha256: the sha256 of the artifact in Pulp to create a package from
        filename: the full filename of the package to create
        repository_pk: the optional pk of the repository to add the content to
    """
    pre_check = PythonPackageContent.objects.filter(sha256=artifact_sha256)
    content_to_add = pre_check or create_content(artifact_sha256, filename)
    content_to_add.get().touch()
    if repository_pk:
        repository = PythonRepository.objects.get(pk=repository_pk)
        with repository.new_version() as new_version:
            new_version.add_content(content_to_add)


def upload_group(session_pk, repository_pk=None):
    """
    Uploads a Python Package to Pulp

    Args:
        session_pk: the session that has the artifacts to upload
        repository_pk: optional repository to add Content to
    """
    s_query = Session.objects.select_for_update().filter(pk=session_pk)
    while True:
        with transaction.atomic():
            session_data = s_query.first().get_decoded()
            now = datetime.now(tz=timezone.utc)
            start_time = datetime.fromisoformat(session_data['start'])
            if now >= start_time:
                content_to_add = PythonPackageContent.objects.none()
                for artifact_sha256, filename in session_data['artifacts']:
                    pre_check = PythonPackageContent.objects.filter(sha256=artifact_sha256)
                    content = pre_check or create_content(artifact_sha256, filename)
                    content.get().touch()
                    content_to_add |= content

                if repository_pk:
                    repository = PythonRepository.objects.get(pk=repository_pk)
                    with repository.new_version() as new_version:
                        new_version.add_content(content_to_add)
                return
            else:
                sleep_time = start_time - now
        time.sleep(sleep_time.seconds)


def create_content(artifact_sha256, filename):
    """
    Creates PythonPackageContent from artifact.

    Args:
        artifact_sha256: validated artifact
        filename: file name
    Returns:
        queryset of the new created content
    """
    artifact = Artifact.objects.get(sha256=artifact_sha256)
    metadata = get_project_metadata_from_artifact(filename, artifact)

    data = parse_project_metadata(vars(metadata))
    data['packagetype'] = metadata.packagetype
    data['version'] = metadata.version
    data['filename'] = filename
    data['sha256'] = artifact.sha256

    @transaction.atomic()
    def create():
        content = PythonPackageContent.objects.create(**data)
        ContentArtifact.objects.create(
            artifact=artifact, content=content, relative_path=filename
        )
        return content

    new_content = create()
    resource = CreatedResource(content_object=new_content)
    resource.save()

    return PythonPackageContent.objects.filter(pk=new_content.pk)



# Generated by Django 2.2.24 on 2021-06-10 14:50

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('python', '0008_pythonpackagecontent_unique_sha256'),
    ]

    operations = [
        migrations.AddField(
            model_name='pythondistribution',
            name='allow_uploads',
            field=models.BooleanField(default=True),
        ),
    ]



# Generated by Django 3.2.6 on 2021-08-24 19:19

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('python', '0009_pythondistribution_allow_uploads'),
    ]

    operations = [
        migrations.AlterField(
            model_name='pythonpackagecontent',
            name='classifiers',
            field=models.JSONField(default=list),
        ),
        migrations.AlterField(
            model_name='pythonpackagecontent',
            name='obsoletes_dist',
            field=models.JSONField(default=list),
        ),
        migrations.AlterField(
            model_name='pythonpackagecontent',
            name='project_urls',
            field=models.JSONField(default=dict),
        ),
        migrations.AlterField(
            model_name='pythonpackagecontent',
            name='provides_dist',
            field=models.JSONField(default=list),
        ),
        migrations.AlterField(
            model_name='pythonpackagecontent',
            name='requires_dist',
            field=models.JSONField(default=list),
        ),
        migrations.AlterField(
            model_name='pythonpackagecontent',
            name='requires_external',
            field=models.JSONField(default=list),
        ),
        migrations.AlterField(
            model_name='pythonremote',
            name='excludes',
            field=models.JSONField(default=list),
        ),
        migrations.AlterField(
            model_name='pythonremote',
            name='includes',
            field=models.JSONField(default=list),
        ),
    ]



import pytest

from pulp_smash.pulp3.utils import gen_distribution, gen_repo
from pulp_python.tests.functional.utils import gen_python_remote

from pulpcore.client.pulp_python import (
    ApiClient,
    ContentPackagesApi,
    DistributionsPypiApi,
    PublicationsPypiApi,
    RepositoriesPythonApi,
    RemotesPythonApi,
)


# Bindings API Fixtures

@pytest.fixture
def python_bindings_client(cid, bindings_cfg):
    """Provides the python bindings client object."""
    api_client = ApiClient(bindings_cfg)
    api_client.default_headers["Correlation-ID"] = cid
    return api_client


@pytest.fixture
def python_repo_api_client(python_bindings_client):
    """Provides the Python Repository API client object."""
    return RepositoriesPythonApi(python_bindings_client)


@pytest.fixture
def python_distro_api_client(python_bindings_client):
    """Provides the Python Distribution API client object."""
    return DistributionsPypiApi(python_bindings_client)


@pytest.fixture
def python_content_api_client(python_bindings_client):
    """Provides the Python Package Content API client object."""
    return ContentPackagesApi(python_bindings_client)


@pytest.fixture
def python_remote_api_client(python_bindings_client):
    """Provides the Python Remotes API client object."""
    return RemotesPythonApi(python_bindings_client)


@pytest.fixture
def python_publication_api_client(python_bindings_client):
    """Proves the Python Publication API client object."""
    return PublicationsPypiApi(python_bindings_client)


# Object Generation Fixtures

@pytest.fixture
def python_repo(python_repo_api_client, gen_object_with_cleanup):
    """Creates a Python Repository and deletes it at test cleanup time."""
    return gen_object_with_cleanup(python_repo_api_client, gen_repo())


@pytest.fixture
def python_distribution_factory(python_distro_api_client, gen_object_with_cleanup):
    """A factory to generate a Python Distribution with auto-cleanup."""
    def _gen_python_distribution(**kwargs):
        distro_data = gen_distribution(**kwargs)
        return gen_object_with_cleanup(python_distro_api_client, distro_data)

    yield _gen_python_distribution


@pytest.fixture
def python_publication_factory(python_publication_api_client, gen_object_with_cleanup):
    """A factory to generate a Python Publication with auto-cleanup."""
    def _gen_python_publication(repository, version=None):
        if version:
            body = {"repository_version": f"{repository.versions_href}{version}/"}
        else:
            body = {"repository": repository.pulp_href}
        return gen_object_with_cleanup(python_publication_api_client, body)

    yield _gen_python_publication


@pytest.fixture
def python_remote_factory(python_remote_api_client, gen_object_with_cleanup):
    """A factory to generate a Python Remote with auto-cleanup."""
    def _gen_python_remote(**kwargs):
        body = gen_python_remote(**kwargs)
        return gen_object_with_cleanup(python_remote_api_client, body)

    yield _gen_python_remote



"""Tests that publish file plugin repositories."""
import unittest
from random import choice

from pulp_smash import config
from pulp_smash.pulp3.bindings import monitor_task
from pulp_smash.pulp3.utils import (
    gen_repo,
    get_content,
    gen_distribution,
    get_versions,
)

from pulp_python.tests.functional.constants import PYTHON_CONTENT_NAME
from pulp_python.tests.functional.utils import (
    gen_python_client,
    gen_python_remote,
)
from pulp_python.tests.functional.utils import set_up_module as setUpModule  # noqa:F401

from pulpcore.client.pulp_python import (
    DistributionsPypiApi,
    PublicationsPypiApi,
    RepositoryAddRemoveContent,
    RepositoriesPythonApi,
    RepositorySyncURL,
    RemotesPythonApi,
    PythonPythonPublication,
)
from pulpcore.client.pulp_python.exceptions import ApiException


class PublishAnyRepoVersionTestCase(unittest.TestCase):
    """Test whether a particular repository version can be published.

    This test targets the following issues:

    * `Pulp #3324 <https://pulp.plan.io/issues/3324>`_
    * `Pulp Smash #897 <https://github.com/pulp/pulp-smash/issues/897>`_
    """

    @classmethod
    def setUpClass(cls):
        """Create class-wide variables."""
        cls.cfg = config.get_config()

        client = gen_python_client()
        cls.repo_api = RepositoriesPythonApi(client)
        cls.remote_api = RemotesPythonApi(client)
        cls.publications = PublicationsPypiApi(client)
        cls.distributions = DistributionsPypiApi(client)

    def setUp(self):
        """Create a new repository before each test."""
        body = gen_python_remote()
        remote = self.remote_api.create(body)

        repo = self.repo_api.create(gen_repo())

        repository_sync_data = RepositorySyncURL(remote=remote.pulp_href)
        sync_response = self.repo_api.sync(repo.pulp_href, repository_sync_data)
        monitor_task(sync_response.task)

        self.repo = self.repo_api.read(repo.pulp_href)

    def test_all(self):
        """Test whether a particular repository version can be published.

        1. Create a repository with at least 2 repository versions.
        2. Create a publication by supplying the latest ``repository_version``.
        3. Assert that the publication ``repository_version`` attribute points
           to the latest repository version.
        4. Create a publication by supplying the non-latest ``repository_version``.
        5. Create distribution.
        6. Assert that the publication ``repository_version`` attribute points
           to the supplied repository version.
        7. Assert that an exception is raised when providing two different
           repository versions to be published at same time.
        """
        # Step 1
        repo_content = get_content(self.repo.to_dict())[PYTHON_CONTENT_NAME][:-1]
        for file_content in repo_content:
            repository_modify_data = RepositoryAddRemoveContent(
                remove_content_units=[file_content["pulp_href"]]
            )
            modify_response = self.repo_api.modify(self.repo.pulp_href, repository_modify_data)
            monitor_task(modify_response.task)
        version_hrefs = tuple(ver["pulp_href"] for ver in get_versions(self.repo.to_dict()))
        non_latest = choice(version_hrefs[1:-1])

        # Step 2
        publish_data = PythonPythonPublication(repository=self.repo.pulp_href)
        publication = self.create_publication(publish_data)

        # Step 3
        self.assertEqual(publication.repository_version, version_hrefs[-1])

        # Step 4
        publish_data = PythonPythonPublication(repository_version=non_latest)
        publication = self.create_publication(publish_data)

        # Step 5
        body = gen_distribution()
        body["base_path"] = "pulp_pre_upgrade_test"
        body["publication"] = publication.pulp_href

        distribution_response = self.distributions.create(body)
        created_resources = monitor_task(distribution_response.task).created_resources
        distribution = self.distributions.read(created_resources[0])

        # Step 6
        self.assertEqual(publication.repository_version, non_latest)

        # Step 7
        with self.assertRaises(ApiException):
            body = {"repository": self.repo.pulp_href, "repository_version": non_latest}
            self.publications.create(body)

        # Step 8
        url = self.cfg.get_content_host_base_url() + "/pulp/content/pulp_pre_upgrade_test/"
        self.assertEqual(url, distribution.base_url, url)

    def create_publication(self, publish_data):
        """Create a new publication from the passed data."""
        publish_response = self.publications.create(publish_data)
        created_resources = monitor_task(publish_response.task).created_resources
        publication_href = created_resources[0]
        return self.publications.read(publication_href)



import logging
from gettext import gettext as _

from rest_framework import serializers
from pulp_python.app.utils import DIST_EXTENSIONS
from pulpcore.plugin.models import Artifact
from django.db.utils import IntegrityError

log = logging.getLogger(__name__)


class SummarySerializer(serializers.Serializer):
    """
    A Serializer for summary information of an index.
    """

    projects = serializers.IntegerField(help_text=_("Number of Python projects in index"))
    releases = serializers.IntegerField(
        help_text=_("Number of Python distribution releases in index")
    )
    files = serializers.IntegerField(help_text=_("Number of files for all distributions in index"))


class PackageMetadataSerializer(serializers.Serializer):
    """
    A Serializer for a package's metadata.
    """

    last_serial = serializers.IntegerField(help_text=_("Cache value from last PyPI sync"))
    info = serializers.JSONField(help_text=_("Core metadata of the package"))
    releases = serializers.JSONField(help_text=_("List of all the releases of the package"))
    urls = serializers.JSONField()


class PackageUploadSerializer(serializers.Serializer):
    """
    A Serializer for Python packages being uploaded to the index.
    """

    content = serializers.FileField(
        help_text=_("A Python package release file to upload to the index."),
        required=True,
        write_only=True,
    )
    action = serializers.CharField(
        help_text=_("Defaults to `file_upload`, don't change it or request will fail!"),
        default="file_upload",
        source=":action"
    )
    sha256_digest = serializers.CharField(
        help_text=_("SHA256 of package to validate upload integrity."),
        required=True,
        min_length=64,
        max_length=64,
    )

    def validate(self, data):
        """Validates the request."""
        action = data.get(":action")
        if action != "file_upload":
            raise serializers.ValidationError(
                _("We do not support the :action {}").format(action)
            )
        file = data.get("content")
        for ext, packagetype in DIST_EXTENSIONS.items():
            if file.name.endswith(ext):
                break
        else:
            raise serializers.ValidationError(_(
                "Extension on {} is not a valid python extension "
                "(.whl, .exe, .egg, .tar.gz, .tar.bz2, .zip)").format(file.name)
            )
        sha256 = data.get("sha256_digest")
        digests = {"sha256": sha256} if sha256 else None
        artifact = Artifact.init_and_validate(file, expected_digests=digests)
        try:
            artifact.save()
        except IntegrityError:
            artifact = Artifact.objects.get(sha256=artifact.sha256)
            artifact.touch()
            log.info(f"Artifact for {file.name} already existed in database")
        data["content"] = (artifact, file.name)
        return data


class PackageUploadTaskSerializer(serializers.Serializer):
    """
    A Serializer for responding to a package upload task.
    """

    session = serializers.CharField()
    task = serializers.CharField()
    task_start_time = serializers.DateTimeField()



# coding=utf-8
"""Tests that verify download of content served by Pulp."""
import hashlib
from random import choice
from urllib.parse import urljoin

from pulp_smash import utils
from pulp_smash.pulp3.utils import (
    download_content_unit,
    get_content_summary,
)
from pulp_python.tests.functional.constants import (
    PYTHON_FIXTURE_URL,
    PYTHON_MD_PROJECT_SPECIFIER,
    PYTHON_MD_FIXTURE_SUMMARY,
    PYTHON_LG_FIXTURE_SUMMARY,
    PYTHON_LG_PROJECT_SPECIFIER,
)
from pulp_python.tests.functional.utils import (
    cfg,
    get_python_content_paths,
    TestCaseUsingBindings,
    TestHelpersMixin,
)
from pulp_python.tests.functional.utils import set_up_module as setUpModule  # noqa:F401


class DownloadContentTestCase(TestCaseUsingBindings, TestHelpersMixin):
    """Verify whether content served by pulp can be downloaded."""

    def test_all(self):
        """Verify whether content served by pulp can be downloaded.

        The process of publishing content is more involved in Pulp 3 than it
        was under Pulp 2. Given a repository, the process is as follows:

        1. Create a publication from the repository. (The latest repository
           version is selected if no version is specified.) A publication is a
           repository version plus metadata.
        2. Create a distribution from the publication. The distribution defines
           at which URLs a publication is available, e.g.
           ``http://example.com/content/foo/`` and
           ``http://example.com/content/bar/``.

        Do the following:

        1. Create, populate, publish, and distribute a repository.
        2. Select a random content unit in the distribution. Download that
           content unit from Pulp, and verify that the content unit has the
           same checksum when fetched directly from Pulp-Fixtures.

        This test targets the following issues:

        * `Pulp #2895 <https://pulp.plan.io/issues/2895>`_
        * `Pulp Smash #872 <https://github.com/pulp/pulp-smash/issues/872>`_
        """
        remote = self._create_remote()
        repo = self._create_repo_and_sync_with_remote(remote)
        pub = self._create_publication(repo)
        distro = self._create_distribution_from_publication(pub)
        # Pick a content unit (of each type), and download it from both Pulp Fixtures…
        unit_paths = [
            choice(paths) for paths in get_python_content_paths(repo.to_dict()).values()
        ]
        fixtures_hashes = [
            hashlib.sha256(
                utils.http_get(
                    urljoin(urljoin(PYTHON_FIXTURE_URL, "packages/"), unit_path[0])
                )
            ).hexdigest()
            for unit_path in unit_paths
        ]

        # …and Pulp.
        pulp_hashes = []
        for unit_path in unit_paths:
            content = download_content_unit(cfg, distro.to_dict(), unit_path[1])
            pulp_hashes.append(hashlib.sha256(content).hexdigest())

        self.assertEqual(fixtures_hashes, pulp_hashes)


class PublishPyPIJSON(TestCaseUsingBindings, TestHelpersMixin):
    """Test whether a distributed Python repository has a PyPI json endpoint
    a.k.a Can be consumed by another Pulp instance

    Test targets the following issue:

    * `Pulp #2886 <https://pulp.plan.io/issues/2886>`_
    """

    def test_basic_pulp_to_pulp_sync(self):
        """
        This test checks that the JSON endpoint is setup correctly to allow one Pulp instance
        to perform a basic sync from another Pulp instance
        """
        body = {"includes": PYTHON_LG_PROJECT_SPECIFIER, "prereleases": True}
        remote = self._create_remote(**body)
        repo = self._create_repo_and_sync_with_remote(remote)
        pub = self._create_publication(repo)
        distro = self._create_distribution_from_publication(pub)
        url_fragments = [
            cfg.get_content_host_base_url(),
            "pulp/content",
            distro.base_path,
            ""
        ]
        unit_url = "/".join(url_fragments)

        # Sync using old Pulp content api
        body["url"] = unit_url
        remote = self._create_remote(**body)
        repo2 = self._create_repo_and_sync_with_remote(remote)
        self.assertEqual(get_content_summary(repo2.to_dict()), PYTHON_LG_FIXTURE_SUMMARY)

        # Sync using new PyPI endpoints
        body["url"] = distro.base_url
        remote = self._create_remote(**body)
        repo3 = self._create_repo_and_sync_with_remote(remote)
        self.assertEqual(get_content_summary(repo3.to_dict()), PYTHON_LG_FIXTURE_SUMMARY)

    def test_full_fixtures_to_pulp_sync(self):
        """
        This test checks that Pulp can fully sync another Python Package repository that is not
        PyPI. This reads the repository's simple page if XMLRPC isn't supported.
        """
        remote = self._create_remote(includes="", prereleases=True)
        repo = self._create_repo_and_sync_with_remote(remote)
        self.assertEqual(get_content_summary(repo.to_dict()), PYTHON_LG_FIXTURE_SUMMARY)

    def test_full_pulp_to_pulp_sync(self):
        """
        This test checks that Pulp can fully sync all packages from another Pulp instance
        without having to specify the includes field.
        """
        remote = self._create_remote(includes=PYTHON_MD_PROJECT_SPECIFIER)
        repo = self._create_repo_and_sync_with_remote(remote)
        # Test using live generated simple pages
        distro = self._create_distribution_from_repo(repo)

        remote = self._create_remote(includes="", url=distro.base_url)
        repo2 = self._create_repo_and_sync_with_remote(remote)
        self.assertEqual(get_content_summary(repo2.to_dict()), PYTHON_MD_FIXTURE_SUMMARY)

        # Now test using publication simple pages
        pub = self._create_publication(repo)
        distro2 = self._create_distribution_from_publication(pub)
        remote = self._create_remote(includes="", url=distro2.base_url, prereleases=True)

        repo3 = self._create_repo_and_sync_with_remote(remote)
        self.assertEqual(get_content_summary(repo3.to_dict()), PYTHON_MD_FIXTURE_SUMMARY)



# Generated by Django 2.2.3 on 2020-06-18 13:55

import django.contrib.postgres.fields.jsonb
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('core', '0032_export_to_chunks'),
    ]

    operations = [
        migrations.CreateModel(
            name='PythonPublication',
            fields=[
                ('publication_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, related_name='python_pythonpublication', serialize=False, to='core.Publication')),
            ],
            options={
                'default_related_name': '%(app_label)s_%(model_name)s',
            },
            bases=('core.publication',),
        ),
        migrations.CreateModel(
            name='PythonRemote',
            fields=[
                ('remote_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, related_name='python_pythonremote', serialize=False, to='core.Remote')),
                ('prereleases', models.BooleanField(default=False)),
                ('includes', django.contrib.postgres.fields.jsonb.JSONField(default=list)),
                ('excludes', django.contrib.postgres.fields.jsonb.JSONField(default=list)),
            ],
            options={
                'default_related_name': '%(app_label)s_%(model_name)s',
            },
            bases=('core.remote',),
        ),
        migrations.CreateModel(
            name='PythonRepository',
            fields=[
                ('repository_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, related_name='python_pythonrepository', serialize=False, to='core.Repository')),
            ],
            options={
                'default_related_name': '%(app_label)s_%(model_name)s',
            },
            bases=('core.repository',),
        ),
        migrations.CreateModel(
            name='PythonPackageContent',
            fields=[
                ('content_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, related_name='python_pythonpackagecontent', serialize=False, to='core.Content')),
                ('filename', models.TextField(db_index=True, unique=True)),
                ('packagetype', models.TextField(choices=[('bdist_dmg', 'bdist_dmg'), ('bdist_dumb', 'bdist_dumb'), ('bdist_egg', 'bdist_egg'), ('bdist_msi', 'bdist_msi'), ('bdist_rpm', 'bdist_rpm'), ('bdist_wheel', 'bdist_wheel'), ('bdist_wininst', 'bdist_wininst'), ('sdist', 'sdist')])),
                ('name', models.TextField()),
                ('version', models.TextField()),
                ('metadata_version', models.TextField()),
                ('summary', models.TextField()),
                ('description', models.TextField()),
                ('keywords', models.TextField()),
                ('home_page', models.TextField()),
                ('download_url', models.TextField()),
                ('author', models.TextField()),
                ('author_email', models.TextField()),
                ('maintainer', models.TextField()),
                ('maintainer_email', models.TextField()),
                ('license', models.TextField()),
                ('requires_python', models.TextField()),
                ('project_url', models.TextField()),
                ('platform', models.TextField()),
                ('supported_platform', models.TextField()),
                ('requires_dist', django.contrib.postgres.fields.jsonb.JSONField(default=list)),
                ('provides_dist', django.contrib.postgres.fields.jsonb.JSONField(default=list)),
                ('obsoletes_dist', django.contrib.postgres.fields.jsonb.JSONField(default=list)),
                ('requires_external', django.contrib.postgres.fields.jsonb.JSONField(default=list)),
                ('classifiers', django.contrib.postgres.fields.jsonb.JSONField(default=list)),
            ],
            options={
                'default_related_name': '%(app_label)s_%(model_name)s',
                'unique_together': {('filename',)},
            },
            bases=('core.content',),
        ),
        migrations.CreateModel(
            name='PythonDistribution',
            fields=[
                ('basedistribution_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, related_name='python_pythondistribution', serialize=False, to='core.BaseDistribution')),
                ('publication', models.ForeignKey(null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='python_pythondistribution', to='core.Publication')),
            ],
            options={
                'default_related_name': '%(app_label)s_%(model_name)s',
            },
            bases=('core.basedistribution',),
        ),
    ]



# coding=utf-8
"""Tests that python plugin can fully mirror PyPi and other Pulp repositories"""
import unittest

from pulp_smash import config, cli
from pulp_smash.pulp3.bindings import delete_orphans, monitor_task
from pulp_smash.pulp3.utils import gen_repo, get_content_summary

from pulp_python.tests.functional.constants import (
    PULP_CONTENT_BASE_URL,
    PULP_PYPI_BASE_URL,
    PYTHON_CONTENT_NAME,
    PYPI_URL,
    PYTHON_XS_FIXTURE_CHECKSUMS,
)
from pulp_python.tests.functional.utils import (
    cfg,
    gen_python_client,
    gen_python_remote,
    TestCaseUsingBindings,
    TestHelpersMixin,
)
from pulp_python.tests.functional.utils import set_up_module as setUpModule  # noqa:F401

from pulpcore.client.pulpcore import TasksApi, ApiClient as CoreApiClient, Configuration
from pulpcore.client.pulp_python import (
    RepositoriesPythonApi,
    RepositorySyncURL,
    RemotesPythonApi,
)
from pypi_simple import parse_repo_project_response
import requests
import socket
from urllib.parse import urljoin, urlsplit


@unittest.skip
class PyPiMirrorTestCase(unittest.TestCase):
    """
    Testing Pulp's full syncing ability of Python repositories

     This test targets the following issues:

    * `Pulp #985 <https://pulp.plan.io/issues/985>`_
    """

    @classmethod
    def setUpClass(cls):
        """Create class-wide variables."""
        cls.cfg = config.get_config()
        cls.client = gen_python_client()
        configuration = Configuration()
        configuration.username = 'admin'
        configuration.password = 'password'
        configuration.host = 'http://{}:24817'.format(socket.gethostname())
        configuration.safe_chars_for_path_param = '/'
        cls.core_client = CoreApiClient(configuration)

    def test_on_demand_pypi_full_sync(self):
        """This test syncs all of PyPi"""
        repo_api = RepositoriesPythonApi(self.client)
        remote_api = RemotesPythonApi(self.client)
        tasks_api = TasksApi(self.core_client)

        repo = repo_api.create(gen_repo())
        self.addCleanup(repo_api.delete, repo.pulp_href)

        body = gen_python_remote("https://pypi.org", includes=[], policy="on_demand")
        remote = remote_api.create(body)
        self.addCleanup(remote_api.delete, remote.pulp_href)

        # Sync the repository.
        self.assertEqual(repo.latest_version_href, f"{repo.pulp_href}versions/0/")
        repository_sync_data = RepositorySyncURL(remote=remote.pulp_href)
        sync_response = repo_api.sync(repo.pulp_href, repository_sync_data)
        monitor_task(sync_response.task)
        repo = repo_api.read(repo.pulp_href)

        sync_task = tasks_api.read(sync_response.task)
        time_diff = sync_task.finished_at - sync_task.started_at
        print("Delete time: {} seconds".format(time_diff.seconds))

        self.assertIsNotNone(repo.latest_version_href)
        # As of August 11 2020, all_packages() returns 253,587 packages,
        # only 248,677 of them were downloadable
        self.assertTrue(get_content_summary(repo.to_dict())[PYTHON_CONTENT_NAME] > 245000)


class PullThroughMirror(TestCaseUsingBindings, TestHelpersMixin):
    """
    Testing Pulp's ability to act as a pull-through cache

    This test targets the following issues:

    * `Pulp #381 https://github.com/pulp/pulp_python/issues/381`_
    """

    @classmethod
    def setUpClass(cls):
        """Sets up class variables."""
        super().setUpClass()
        cls.cli_client = cli.Client(cfg)
        cls.HOST = cls.client.configuration.host
        cls.PYPI_HOST = urljoin(cls.HOST, PULP_PYPI_BASE_URL)

    def test_pull_through_install(self):
        """Tests that a pull-through distro can be installed from."""
        remote = self._create_remote(url=PYPI_URL)
        distro = self._create_distribution(remote=remote.pulp_href)

        if self.cli_client.run(("pip", "list")).stdout.find("numpy") == -1:
            content = self.content_api.list(name="numpy")
            self.assertEqual(content.count, 0, msg="numpy content already present in test")
            host = urlsplit(PULP_CONTENT_BASE_URL).hostname
            url = urljoin(self.PYPI_HOST, f"{distro.base_path}/simple/")
            out = self.cli_client.run(
                ("pip", "install", "--trusted-host", host, "-i", url, "numpy")
            )
            self.addCleanup(delete_orphans)
            self.assertTrue(self.cli_client.run(("pip", "list")).stdout.find("numpy") != -1, out)
            self.addCleanup(self.cli_client.run, ("pip", "uninstall", "numpy", "-y"))
            content = self.content_api.list(name="numpy")
            self.assertEqual(content.count, 1)
        else:
            self.skipTest("Uninstall numpy before running this test")

    def test_pull_through_simple(self):
        """Tests that the simple page is properly modified when requesting a pull-through."""
        remote = self._create_remote(url=PYPI_URL)
        distro = self._create_distribution(remote=remote.pulp_href)

        url = urljoin(self.PYPI_HOST, f"{distro.base_path}/simple/shelf-reader/")
        response = requests.get(url)
        project_page = parse_repo_project_response("shelf-reader", response)

        self.assertEqual(len(project_page.packages), 2)
        for package in project_page.packages:
            self.assertIn(package.filename, PYTHON_XS_FIXTURE_CHECKSUMS)
            relative_path = f"{distro.base_path}/{package.filename}?redirect="
            self.assertIn(urljoin(PULP_CONTENT_BASE_URL, relative_path), package.url)
            digests = package.get_digests()
            self.assertEqual(PYTHON_XS_FIXTURE_CHECKSUMS[package.filename], digests["sha256"])

    def test_pull_through_with_repo(self):
        """Tests that if content is already in repository, pull-through isn't used."""
        remote = self._create_remote()
        repo = self._create_repo_and_sync_with_remote(remote)
        self.addCleanup(delete_orphans)
        distro = self._create_distribution(remote=remote.pulp_href, repository=repo.pulp_href)

        url = urljoin(self.PYPI_HOST, f"{distro.base_path}/simple/shelf-reader/")
        response = requests.get(url)
        project_page = parse_repo_project_response("shelf-reader", response)

        self.assertEqual(len(project_page.packages), 2)
        for package in project_page.packages:
            self.assertIn(package.filename, PYTHON_XS_FIXTURE_CHECKSUMS)
            self.assertNotIn("?redirect=", package.url)



"""Tests all the PyPI apis available at `pypi/`."""
import os
import requests
import subprocess
import tempfile

from urllib.parse import urljoin

from pulp_smash.pulp3.bindings import monitor_task, tasks as task_api
from pulp_smash.pulp3.utils import get_added_content_summary, get_content_summary
from pulp_python.tests.functional.constants import (
    PYTHON_CONTENT_NAME,
    PYTHON_SM_PROJECT_SPECIFIER,
    PYTHON_SM_FIXTURE_RELEASES,
    PYTHON_SM_FIXTURE_CHECKSUMS,
    PYTHON_MD_PROJECT_SPECIFIER,
    PYTHON_MD_PYPI_SUMMARY,
    PULP_CONTENT_BASE_URL,
    PULP_PYPI_BASE_URL,
    PYTHON_EGG_FILENAME,
    PYTHON_EGG_URL,
    PYTHON_EGG_SHA256,
    PYTHON_WHEEL_FILENAME,
    PYTHON_WHEEL_URL,
    PYTHON_WHEEL_SHA256,
    SHELF_PYTHON_JSON,
)

from pulp_python.tests.functional.utils import (
    py_client as client,
    ensure_simple,
    TestCaseUsingBindings,
    TestHelpersMixin,
)
from pulpcore.client.pulp_python import PypiApi

PYPI_LAST_SERIAL = "X-PYPI-LAST-SERIAL"
PYPI_SERIAL_CONSTANT = 1000000000
HOST = client.configuration.host
PYPI_HOST = urljoin(HOST, PULP_PYPI_BASE_URL)


class PyPISummaryTestCase(TestCaseUsingBindings, TestHelpersMixin):
    """Tests the summary response of the base url of an index."""

    @classmethod
    def setUpClass(cls):
        """Set up class variables."""
        super().setUpClass()
        cls.pypi_api = PypiApi(client)

    def test_empty_index(self):
        """Checks that summary stats are 0 when index is empty."""
        _, distro = self._create_empty_repo_and_distribution()

        summary = self.pypi_api.read(path=distro.base_path)
        self.assertTrue(not any(summary.to_dict().values()))

    def test_live_index(self):
        """Checks summary stats are correct for indexes pointing to repositories."""
        remote = self._create_remote(includes=PYTHON_MD_PROJECT_SPECIFIER)
        repo = self._create_repo_and_sync_with_remote(remote)
        distro = self._create_distribution_from_repo(repo)

        summary = self.pypi_api.read(path=distro.base_path)
        self.assertDictEqual(summary.to_dict(), PYTHON_MD_PYPI_SUMMARY)

    def test_published_index(self):
        """Checks summary stats are correct for indexes pointing to publications."""
        remote = self._create_remote(includes=PYTHON_MD_PROJECT_SPECIFIER)
        repo = self._create_repo_and_sync_with_remote(remote)
        pub = self._create_publication(repo)
        distro = self._create_distribution_from_publication(pub)

        summary = self.pypi_api.read(path=distro.base_path)
        self.assertDictEqual(summary.to_dict(), PYTHON_MD_PYPI_SUMMARY)


class PyPIPackageUpload(TestCaseUsingBindings, TestHelpersMixin):
    """Tests the package upload endpoints of an index."""

    @classmethod
    def setUpClass(cls):
        """Set up class variables."""
        super().setUpClass()
        cls.pypi_api = PypiApi(client)
        cls.dists_dir = tempfile.TemporaryDirectory()
        cls.egg = os.path.join(cls.dists_dir.name, PYTHON_EGG_FILENAME)
        cls.wheel = os.path.join(cls.dists_dir.name, PYTHON_WHEEL_FILENAME)
        with open(cls.egg, "wb") as fp:
            fp.write(requests.get(PYTHON_EGG_URL).content)
        with open(cls.wheel, "wb") as fp:
            fp.write(requests.get(PYTHON_WHEEL_URL).content)

    @classmethod
    def tearDownClass(cls):
        """Tear down class variables."""
        cls.dists_dir.cleanup()

    def test_package_upload(self):
        """Tests that packages can be uploaded."""
        repo, distro = self._create_empty_repo_and_distribution()
        url = urljoin(PYPI_HOST, distro.base_path + "/legacy/")
        response = requests.post(
            url,
            data={"sha256_digest": PYTHON_EGG_SHA256},
            files={"content": open(self.egg, "rb")},
        )
        self.assertEqual(response.status_code, 200)
        task = response.json()["task"]
        monitor_task(task)
        content = get_added_content_summary(repo, f"{repo.versions_href}1/")
        self.assertDictEqual({PYTHON_CONTENT_NAME: 1}, content)
        # Test re-uploading same package gives 400 Bad Request
        response = requests.post(
            url,
            data={"sha256_digest": PYTHON_EGG_SHA256},
            files={"content": open(self.egg, "rb")},
        )
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.reason, f"Package {PYTHON_EGG_FILENAME} already exists in index"
        )

    def test_package_upload_session(self):
        """Tests that multiple uploads will be broken up into multiple tasks."""
        repo, distro = self._create_empty_repo_and_distribution()
        url = urljoin(PYPI_HOST, distro.base_path + "/legacy/")
        session = requests.Session()
        response = session.post(
            url,
            data={"sha256_digest": PYTHON_EGG_SHA256},
            files={"content": open(self.egg, "rb")},
        )
        self.assertEqual(response.status_code, 200)
        response = response.json()
        monitor_task(response["task"])
        response2 = session.post(
            url,
            data={"sha256_digest": PYTHON_WHEEL_SHA256},
            files={"content": open(self.wheel, "rb")},
        )
        self.assertEqual(response2.status_code, 200)
        response2 = response2.json()
        self.assertNotEqual(response["task"], response2["task"])
        monitor_task(response2["task"])
        content = get_content_summary(repo, f"{repo.versions_href}2/")
        self.assertDictEqual({PYTHON_CONTENT_NAME: 2}, content)

    def test_package_upload_simple(self):
        """Tests that the package upload endpoint exposed at `/simple/` works."""
        repo, distro = self._create_empty_repo_and_distribution()
        url = urljoin(PYPI_HOST, distro.base_path + "/simple/")
        response = requests.post(
            url,
            data={"sha256_digest": PYTHON_EGG_SHA256},
            files={"content": open(self.egg, "rb")},
        )
        self.assertEqual(response.status_code, 200)
        task = response.json()["task"]
        monitor_task(task)
        content = get_added_content_summary(repo, f"{repo.versions_href}1/")
        self.assertDictEqual({PYTHON_CONTENT_NAME: 1}, content)

    def test_twine_upload(self):
        """Tests that packages can be properly uploaded through Twine."""
        repo, distro = self._create_empty_repo_and_distribution()
        url = urljoin(PYPI_HOST, distro.base_path + "/legacy/")
        username, password = "admin", "password"
        subprocess.run(
            (
                "twine",
                "upload",
                "--repository-url",
                url,
                self.dists_dir.name + "/*",
                "-u",
                username,
                "-p",
                password,
            ),
            capture_output=True,
            check=True,
        )
        tasks = task_api.list(reserved_resources_record=[repo.pulp_href]).results
        for task in reversed(tasks):
            t = monitor_task(task.pulp_href)
            repo_ver_href = t.created_resources[-1]
        content = get_content_summary(repo, f"{repo_ver_href}")
        self.assertDictEqual({PYTHON_CONTENT_NAME: 2}, content)

        # Test re-uploading same packages gives error
        with self.assertRaises(subprocess.CalledProcessError):
            subprocess.run(
                (
                    "twine",
                    "upload",
                    "--repository-url",
                    url,
                    self.dists_dir.name + "/*",
                    "-u",
                    username,
                    "-p",
                    password,
                ),
                capture_output=True,
                check=True,
            )

        # Test re-uploading same packages with --skip-existing works
        output = subprocess.run(
            (
                "twine",
                "upload",
                "--repository-url",
                url,
                self.dists_dir.name + "/*",
                "-u",
                username,
                "-p",
                password,
                "--skip-existing",
            ),
            capture_output=True,
            check=True,
            text=True
        )
        self.assertEqual(output.stdout.count("Skipping"), 2)


class PyPISimpleApi(TestCaseUsingBindings, TestHelpersMixin):
    """Tests that the simple api is correct."""

    def test_simple_redirect_with_publications(self):
        """Checks that requests to `/simple/` get redirected when serving a publication."""
        remote = self._create_remote()
        repo = self._create_repo_and_sync_with_remote(remote)
        pub = self._create_publication(repo)
        distro = self._create_distribution_from_publication(pub)
        response = requests.get(urljoin(PYPI_HOST, f"{distro.base_path}/simple/"))
        self.assertEqual(
            response.url,
            str(urljoin(PULP_CONTENT_BASE_URL, f"{distro.base_path}/simple/")),
        )

    def test_simple_correctness_live(self):
        """Checks that the simple api on live distributions are correct."""
        remote = self._create_remote(includes=PYTHON_SM_PROJECT_SPECIFIER)
        repo = self._create_repo_and_sync_with_remote(remote)
        distro = self._create_distribution_from_repo(repo)
        proper, msgs = ensure_simple(
            urljoin(PYPI_HOST, f"{distro.base_path}/simple/"),
            PYTHON_SM_FIXTURE_RELEASES,
            sha_digests=PYTHON_SM_FIXTURE_CHECKSUMS,
        )
        self.assertTrue(proper, msg=msgs)


class PyPIPackageMetadata(TestCaseUsingBindings, TestHelpersMixin):
    """Test whether a distributed Python repository has a PyPI json endpoint."""

    def test_pypi_json(self):
        """Checks the data of `pypi/{package_name}/json` endpoint."""
        remote = self._create_remote(policy="immediate")
        repo = self._create_repo_and_sync_with_remote(remote)
        distro = self._create_distribution_from_repo(repo)
        rel_url = f"{distro.base_path}/pypi/shelf-reader/json"
        response = requests.get(urljoin(PYPI_HOST, rel_url))
        self.assert_pypi_json(response.json())

    def test_pypi_json_content_app(self):
        """Checks that the pypi json endpoint of the content app still works. Needs Publication."""
        remote = self._create_remote(policy="immediate")
        repo = self._create_repo_and_sync_with_remote(remote)
        pub = self._create_publication(repo)
        distro = self._create_distribution_from_publication(pub)
        rel_url = f"{distro.base_path}/pypi/shelf-reader/json/"
        response = requests.get(urljoin(PULP_CONTENT_BASE_URL, rel_url))
        self.assert_pypi_json(response.json())

    def test_pypi_last_serial(self):
        """
        Checks that the endpoint has the header PYPI_LAST_SERIAL and is set
        TODO when serial field is added to Repo's, check this header against that
        """
        remote = self._create_remote()
        repo = self._create_repo_and_sync_with_remote(remote)
        pub = self._create_publication(repo)
        distro = self._create_distribution_from_publication(pub)
        content_url = urljoin(
            PULP_CONTENT_BASE_URL, f"{distro.base_path}/pypi/shelf-reader/json"
        )
        pypi_url = urljoin(PYPI_HOST, f"{distro.base_path}/pypi/shelf-reader/json/")
        for url in [content_url, pypi_url]:
            response = requests.get(url)
            self.assertIn(PYPI_LAST_SERIAL, response.headers, msg=url)
            self.assertEqual(
                response.headers[PYPI_LAST_SERIAL], str(PYPI_SERIAL_CONSTANT), msg=url
            )

    def assert_pypi_json(self, package):
        """Asserts that shelf-reader package json is correct."""
        self.assertEqual(SHELF_PYTHON_JSON["last_serial"], package["last_serial"])
        self.assertTrue(SHELF_PYTHON_JSON["info"].items() <= package["info"].items())
        self.assertEqual(len(SHELF_PYTHON_JSON["urls"]), len(package["urls"]))
        self.assert_download_info(
            SHELF_PYTHON_JSON["urls"], package["urls"], "Failed to match URLS"
        )
        self.assertTrue(
            SHELF_PYTHON_JSON["releases"].keys() <= package["releases"].keys()
        )
        for version in SHELF_PYTHON_JSON["releases"].keys():
            self.assert_download_info(
                SHELF_PYTHON_JSON["releases"][version],
                package["releases"][version],
                "Failed to match version",
            )

    def assert_download_info(self, expected, received, message="Failed to match"):
        """
        Each version has a list of dists of that version, but the lists might
        not be in the same order, so check each dist of the second list
        """
        for dist in expected:
            dist = dict(dist)
            matched = False
            dist_items = dist.items()
            for dist2 in received:
                dist2 = dict(dist2)
                dist2["digests"].pop("md5", "")
                if dist_items <= dist2.items():
                    matched = True
                    break
            self.assertTrue(matched, message)



import logging
import requests

from rest_framework.viewsets import ViewSet
from rest_framework.response import Response
from rest_framework.permissions import IsAuthenticatedOrReadOnly
from django.core.exceptions import ObjectDoesNotExist
from django.shortcuts import redirect
from datetime import datetime, timezone, timedelta

from rest_framework.reverse import reverse
from django.contrib.sessions.models import Session
from django.db import transaction
from django.db.utils import DatabaseError
from django.http.response import (
    Http404,
    HttpResponseForbidden,
    HttpResponseBadRequest,
    StreamingHttpResponse
)
from drf_spectacular.utils import extend_schema
from dynaconf import settings
from itertools import chain
from packaging.utils import canonicalize_name
from urllib.parse import urljoin, urlparse, urlunsplit
from pathlib import PurePath
from pypi_simple.parse_stream import parse_links_stream_response

from pulpcore.plugin.tasking import dispatch
from pulp_python.app.models import (
    PythonDistribution,
    PythonPackageContent,
    PythonPublication,
)
from pulp_python.app.pypi.serializers import (
    SummarySerializer,
    PackageMetadataSerializer,
    PackageUploadSerializer,
    PackageUploadTaskSerializer
)
from pulp_python.app.utils import (
    write_simple_index,
    write_simple_detail,
    python_content_to_json,
    PYPI_LAST_SERIAL,
    PYPI_SERIAL_CONSTANT,
)

from pulp_python.app import tasks

log = logging.getLogger(__name__)

BASE_CONTENT_URL = urljoin(settings.CONTENT_ORIGIN, settings.CONTENT_PATH_PREFIX)


class PyPIMixin:
    """Mixin to get index specific info."""

    @staticmethod
    def get_repository_version(distribution):
        """Finds the repository version this distribution is serving."""
        pub = distribution.publication
        rep = distribution.repository
        if pub:
            return pub.repository_version or pub.repository.latest_version()
        elif rep:
            return rep.latest_version()
        else:
            raise Http404("No repository associated with this index")

    @staticmethod
    def get_distribution(path):
        """Finds the distribution associated with this base_path."""
        distro_qs = PythonDistribution.objects.select_related(
            "repository", "publication", "publication__repository_version", "remote"
        )
        try:
            return distro_qs.get(base_path=path)
        except ObjectDoesNotExist:
            raise Http404(f"No PythonDistribution found for base_path {path}")

    @staticmethod
    def get_content(repository_version):
        """Returns queryset of the content in this repository version."""
        return PythonPackageContent.objects.filter(pk__in=repository_version.content)

    def should_redirect(self, distro, repo_version=None):
        """Checks if there is a publication the content app can serve."""
        if distro.publication:
            return True
        rv = repo_version or self.get_repository_version(distro)
        return PythonPublication.objects.filter(repository_version=rv).exists()

    def get_drvc(self, path):
        """Takes the base_path and returns the distribution, repository_version and content."""
        distro = self.get_distribution(path)
        if distro.remote and not distro.repository and not distro.publication:
            return distro, None, None
        repo_ver = self.get_repository_version(distro)
        content = self.get_content(repo_ver)
        return distro, repo_ver, content


class PackageUploadMixin(PyPIMixin):
    """A Mixin to provide package upload support."""

    def upload(self, request, path):
        """
        Upload a package to the index.

        0. Check if the index allows uploaded packages (live-api-enabled)
        1. Check request is in correct format
        2. Check if the package is in the repository already
        3. If present then reject request
        4. Spawn task to add content if no/old session present
        5. Add uploads to current session to group into one task
        """
        distro = self.get_distribution(path)
        if not distro.allow_uploads:
            return HttpResponseForbidden(reason="Index is not allowing uploads")

        repo = distro.repository
        if not repo:
            return HttpResponseBadRequest(reason="Index is not pointing to a repository")

        serializer = PackageUploadSerializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        artifact, filename = serializer.validated_data["content"]
        repo_content = self.get_content(self.get_repository_version(distro))
        if repo_content.filter(filename=filename).exists():
            return HttpResponseBadRequest(reason=f"Package {filename} already exists in index")

        if settings.PYTHON_GROUP_UPLOADS:
            return self.upload_package_group(repo, artifact, filename, request.session)

        result = dispatch(tasks.upload, exclusive_resources=[artifact, repo],
                          kwargs={"artifact_sha256": artifact.sha256,
                                  "filename": filename,
                                  "repository_pk": str(repo.pk)})
        return Response(data={"task": reverse('tasks-detail', args=[result.pk], request=None)})

    def upload_package_group(self, repo, artifact, filename, session):
        """Steps 4 & 5, spawns tasks to add packages to index."""
        start_time = datetime.now(tz=timezone.utc) + timedelta(seconds=5)
        task = "updated"
        if not session.get("start"):
            task = self.create_group_upload_task(session, repo, artifact, filename, start_time)
        else:
            sq = Session.objects.select_for_update(nowait=True).filter(pk=session.session_key)
            try:
                with transaction.atomic():
                    sq.first()
                    current_start = datetime.fromisoformat(session['start'])
                    if current_start >= datetime.now(tz=timezone.utc):
                        session['artifacts'].append((str(artifact.sha256), filename))
                        session['start'] = str(start_time)
                        session.modified = False
                        session.save()
                    else:
                        raise DatabaseError
            except DatabaseError:
                session.cycle_key()
                task = self.create_group_upload_task(session, repo, artifact, filename, start_time)
        data = {"session": session.session_key, "task": task, "task_start_time": start_time}
        return Response(data=data)

    def create_group_upload_task(self, cur_session, repository, artifact, filename, start_time):
        """Creates the actual task that adds the packages to the index."""
        cur_session['start'] = str(start_time)
        cur_session['artifacts'] = [(str(artifact.sha256), filename)]
        cur_session.modified = False
        cur_session.save()
        result = dispatch(tasks.upload_group, exclusive_resources=[artifact, repository],
                          kwargs={"session_pk": str(cur_session.session_key),
                                  "repository_pk": str(repository.pk)})
        return reverse('tasks-detail', args=[result.pk], request=None)


class SimpleView(ViewSet, PackageUploadMixin):
    """View for the PyPI simple API."""

    permission_classes = [IsAuthenticatedOrReadOnly]

    @extend_schema(summary="Get index simple page")
    def list(self, request, path):
        """Gets the simple api html page for the index."""
        distro, repo_version, content = self.get_drvc(path)
        if self.should_redirect(distro, repo_version=repo_version):
            return redirect(urljoin(BASE_CONTENT_URL, f'{path}/simple/'))
        names = content.order_by('name').values_list('name', flat=True).distinct().iterator()
        return StreamingHttpResponse(write_simple_index(names, streamed=True))

    def pull_through_package_simple(self, package, path, remote):
        """Gets the package's simple page from remote."""
        def parse_url(link):
            parsed = urlparse(link.url)
            digest, _, value = parsed.fragment.partition('=')
            stripped_url = urlunsplit(chain(parsed[:3], ("", "")))
            redirect = f'{path}/{link.text}?redirect={stripped_url}'
            d_url = urljoin(BASE_CONTENT_URL, redirect)
            return link.text, d_url, value if digest == 'sha256' else ''

        url = remote.get_remote_artifact_url(f'simple/{package}/')
        response = requests.get(url, stream=True)
        links = parse_links_stream_response(response)
        packages = (parse_url(link) for link in links)
        return StreamingHttpResponse(write_simple_detail(package, packages, streamed=True))

    @extend_schema(summary="Get package simple page")
    def retrieve(self, request, path, package):
        """Retrieves the simple api html page for a package."""
        distro, repo_ver, content = self.get_drvc(path)
        # Should I redirect if the normalized name is different?
        normalized = canonicalize_name(package)
        if distro.remote:
            if not repo_ver or not content.filter(name__normalize=normalized).exists():
                return self.pull_through_package_simple(normalized, path, distro.remote)
        if self.should_redirect(distro, repo_version=repo_ver):
            return redirect(urljoin(BASE_CONTENT_URL, f'{path}/simple/{normalized}/'))
        packages = (
            content.filter(name__normalize=normalized)
            .values_list('filename', 'sha256', 'name')
            .iterator()
        )
        try:
            present = next(packages)
        except StopIteration:
            raise Http404(f"{normalized} does not exist.")
        else:
            packages = chain([present], packages)
            name = present[2]
        releases = ((f, urljoin(BASE_CONTENT_URL, f'{path}/{f}'), d) for f, d, _ in packages)
        return StreamingHttpResponse(write_simple_detail(name, releases, streamed=True))

    @extend_schema(request=PackageUploadSerializer,
                   responses={200: PackageUploadTaskSerializer},
                   summary="Upload a package")
    def create(self, request, path):
        """
        Upload package to the index.
        This endpoint has the same functionality as the upload endpoint at the `/legacy` url of the
        index. This is provided for convenience for users who want a single index url for all their
        Python tools. (pip, twine, poetry, pipenv, ...)
        """
        return self.upload(request, path)


class MetadataView(ViewSet, PyPIMixin):
    """View for the PyPI JSON metadata endpoint."""

    authentication_classes = []
    permission_classes = []

    @extend_schema(tags=["Pypi: Metadata"],
                   responses={200: PackageMetadataSerializer},
                   summary="Get package metadata")
    def retrieve(self, request, path, meta):
        """
        Retrieves the package's core-metadata specified by
        https://packaging.python.org/specifications/core-metadata/.
        `meta` must be a path in form of `{package}/json/` or `{package}/{version}/json/`
        """
        distro, repo_ver, content = self.get_drvc(path)
        meta_path = PurePath(meta)
        name = None
        version = None
        if meta_path.match("*/*/json"):
            version = meta_path.parts[1]
            name = meta_path.parts[0]
        elif meta_path.match("*/json"):
            name = meta_path.parts[0]
        if name:
            package_content = content.filter(name__iexact=name)
            # TODO Change this value to the Repo's serial value when implemented
            headers = {PYPI_LAST_SERIAL: str(PYPI_SERIAL_CONSTANT)}
            json_body = python_content_to_json(path, package_content, version=version)
            if json_body:
                return Response(data=json_body, headers=headers)
        return Response(status="404")


class PyPIView(ViewSet, PyPIMixin):
    """View for base_url of distribution."""

    authentication_classes = []
    permission_classes = []

    @extend_schema(responses={200: SummarySerializer},
                   summary="Get index summary")
    def retrieve(self, request, path):
        """Gets package summary stats of index."""
        distro, repo_ver, content = self.get_drvc(path)
        files = content.count()
        releases = content.distinct("name", "version").count()
        projects = content.distinct("name").count()
        data = {"projects": projects, "releases": releases, "files": files}
        return Response(data=data)


class UploadView(ViewSet, PackageUploadMixin):
    """View for the `/legacy` upload endpoint."""

    @extend_schema(request=PackageUploadSerializer,
                   responses={200: PackageUploadTaskSerializer},
                   summary="Upload a package")
    def create(self, request, path):
        """
        Upload package to the index.

        This is the endpoint that tools like Twine and Poetry use for their upload commands.
        """
        return self.upload(request, path)



# coding=utf-8
"""Tests that perform actions over content unit."""
from pulp_smash import cli
from pulp_smash.pulp3.bindings import delete_orphans, monitor_task
from pulp_smash.pulp3.utils import modify_repo

from pulp_python.tests.functional.constants import (
    PYTHON_FIXTURE_URL,
    PYTHON_FIXTURES_PACKAGES,
    PYTHON_FIXTURES_FILENAMES,
    PYTHON_LIST_PROJECT_SPECIFIER,
    PYPI_URL,
)

from pulp_python.tests.functional.utils import (
    cfg,
    gen_artifact,
    gen_python_content_attrs,
    TestCaseUsingBindings,
    TestHelpersMixin,
)
from pulp_python.tests.functional.utils import set_up_module as setUpModule  # noqa:F401
from urllib.parse import urljoin, urlsplit

from pulp_smash.utils import http_get


class PipInstallContentTestCase(TestCaseUsingBindings, TestHelpersMixin):
    """
    Verify whether content served by Pulp can be consumed through pip install.
    Workflows tested are:
    1) create repo -> upload package artifact -> create content -> add content to repo -> publish
    repo -> distribute publication -> consume through pip
    2) create repo -> sync with remote -> publish repo -> distribute publication -> consume through
       pip
    """

    @classmethod
    def setUpClass(cls):
        """
        Check if packages to install through tests are already installed
        """
        super().setUpClass()
        cls.cli_client = cli.Client(cfg)
        cls.PACKAGES = PYTHON_FIXTURES_PACKAGES
        cls.PACKAGES_URLS = [
            urljoin(urljoin(PYTHON_FIXTURE_URL, "packages/"), filename)
            for filename in PYTHON_FIXTURES_FILENAMES
        ]
        cls.PACKAGES_FILES = [{"file": http_get(file)} for file in cls.PACKAGES_URLS]
        delete_orphans()
        for pkg in cls.PACKAGES:
            cls.assertFalse(
                cls,
                cls.check_install(cls.cli_client, pkg),
                "{} is already installed".format(pkg),
            )

    def test_workflow_01(self):
        """
        Verify workflow 1
        """
        created_contents = []
        for pkg, filename in zip(self.PACKAGES_URLS, PYTHON_FIXTURES_FILENAMES):
            content_response = self.content_api.create(
                **gen_python_content_attrs(gen_artifact(pkg), filename)
            )
            created_contents.extend(monitor_task(content_response.task).created_resources)
            created_contents = [self.content_api.read(href).to_dict() for href in created_contents]

        repo = self._create_repository()
        # Add content
        modify_repo(cfg, repo.to_dict(), add_units=created_contents)
        repo = self.repo_api.read(repo.pulp_href)
        pub = self._create_publication(repo)
        distro = self._create_distribution_from_publication(pub)

        self.addCleanup(delete_orphans)
        self.check_consume(distro.to_dict())

    def test_workflow_02(self):
        """
        Verify workflow 2

        Do the following:

        1. Create, populate, publish, and distribute a repository.
        2. Pip install a package from the pulp repository.
        3. Check pip install was successful.

        This test targets the following issues:
        * `Pulp #4682 <https://pulp.plan.io/issues/4682>`_
        * `Pulp #4677 <https://pulp.plan.io/issues/4677>`_
        """
        remote = self._create_remote(includes=PYTHON_LIST_PROJECT_SPECIFIER)
        repo = self._create_repo_and_sync_with_remote(remote)
        pub = self._create_publication(repo)
        distro = self._create_distribution_from_publication(pub)
        self.check_consume(distro.to_dict())

    def check_consume(self, distribution):
        """Tests that pip packages hosted in a distribution can be consumed"""
        host_base_url = cfg.get_content_host_base_url()
        url = "".join(
            [host_base_url, "/pulp/content/", distribution["base_path"], "/simple/"]
        )
        for pkg in self.PACKAGES:
            out = self.install(self.cli_client, pkg, host=url)
            self.assertTrue(self.check_install(self.cli_client, pkg), out)
            self.addCleanup(self.uninstall, self.cli_client, pkg)

    @staticmethod
    def check_install(cli_client, package):
        """Returns true if python package is installed, false otherwise"""
        return cli_client.run(("pip", "list")).stdout.find(package) != -1

    @staticmethod
    def install(cli_client, package, host=PYPI_URL):
        """Installs a pip package from the host url"""
        return cli_client.run(
            (
                "pip",
                "install",
                "--no-deps",
                "--trusted-host",
                urlsplit(host).hostname,
                "--trusted-host",
                "ci-azurite",
                "-i",
                host,
                package,
            )
        ).stdout

    @staticmethod
    def uninstall(cli_client, package):
        """
        Uninstalls a pip package and returns the version number
        Uninstall Message format: "Found existing installation: package X.X.X ..."
        """
        return cli_client.run(("pip", "uninstall", package, "-y")).stdout.split()[4]



from pulpcore.plugin import PulpPluginAppConfig


class PulpPythonPluginAppConfig(PulpPluginAppConfig):
    """
    Entry point for pulp_python plugin.
    """

    name = "pulp_python.app"
    label = "python"
    version = "3.7.0"
    python_package_name = "pulp-python"



# Generated by Django 2.2.20 on 2021-04-27 20:14

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('python', '0005_pythonpackagecontent_sha256'),
    ]

    operations = [
        migrations.AddField(
            model_name='pythonrepository',
            name='autopublish',
            field=models.BooleanField(default=False),
        ),
    ]



# Generated by Django 2.2.23 on 2021-06-04 15:50

import django.contrib.postgres.fields.jsonb
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('python', '0006_pythonrepository_autopublish'),
    ]

    operations = [
        migrations.AddField(
            model_name='pythonpackagecontent',
            name='description_content_type',
            field=models.TextField(default=''),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name='pythonpackagecontent',
            name='project_urls',
            field=django.contrib.postgres.fields.jsonb.JSONField(default=dict),
        ),
    ]



import logging

from aiohttp import ClientResponseError, ClientError
from lxml.etree import LxmlError
from gettext import gettext as _
from os import environ

from rest_framework import serializers

from pulpcore.plugin.models import Artifact, ProgressReport, Remote, Repository
from pulpcore.plugin.stages import (
    DeclarativeArtifact,
    DeclarativeContent,
    DeclarativeVersion,
    Stage,
)

from pulp_python.app.models import (
    PythonPackageContent,
    PythonRemote,
)
from pulp_python.app.utils import parse_metadata, PYPI_LAST_SERIAL
from pypi_simple import parse_repo_index_page

from bandersnatch.mirror import Mirror
from bandersnatch.master import Master
from bandersnatch.configuration import BandersnatchConfig
from packaging.requirements import Requirement
from urllib.parse import urljoin

logger = logging.getLogger(__name__)


def sync(remote_pk, repository_pk, mirror):
    """
    Sync content from the remote repository.

    Create a new version of the repository that is synchronized with the remote.

    Args:
        remote_pk (str): The remote PK.
        repository_pk (str): The repository PK.
        mirror (boolean): True for mirror mode, False for additive mode.

    Raises:
        serializers: ValidationError

    """
    remote = PythonRemote.objects.get(pk=remote_pk)
    repository = Repository.objects.get(pk=repository_pk)

    if not remote.url:
        raise serializers.ValidationError(
            detail=_("A remote must have a url attribute to sync.")
        )

    first_stage = PythonBanderStage(remote)
    DeclarativeVersion(first_stage, repository, mirror).create()


def create_bandersnatch_config(remote):
    """Modifies the global Bandersnatch config state for this sync"""
    config = BandersnatchConfig().config
    config["mirror"]["master"] = remote.url
    config["mirror"]["workers"] = str(remote.download_concurrency)
    if not config.has_section("plugins"):
        config.add_section("plugins")
    config["plugins"]["enabled"] = "blocklist_release\n"
    if remote.includes:
        if not config.has_section("allowlist"):
            config.add_section("allowlist")
        config["plugins"]["enabled"] += "allowlist_release\nallowlist_project\n"
        config["allowlist"]["packages"] = "\n".join(remote.includes)
    if remote.excludes:
        if not config.has_section("blocklist"):
            config.add_section("blocklist")
        config["plugins"]["enabled"] += "blocklist_project\n"
        config["blocklist"]["packages"] = "\n".join(remote.excludes)
    if not remote.prereleases:
        config["plugins"]["enabled"] += "prerelease_release\n"
    if remote.package_types:
        rrfm = "regex_release_file_metadata"
        config["plugins"]["enabled"] += rrfm
        if not config.has_section(rrfm):
            config.add_section(rrfm)
        config[rrfm]["any:release_file.packagetype"] = "\n".join(remote.package_types)
    if remote.keep_latest_packages:
        config["plugins"]["enabled"] += "latest_release\n"
        if not config.has_section("latest_release"):
            config.add_section("latest_release")
        config["latest_release"]["keep"] = str(remote.keep_latest_packages)
    if remote.exclude_platforms:
        config["plugins"]["enabled"] += "exclude_platform\n"
        if not config.has_section("blocklist"):
            config.add_section("blocklist")
        config["blocklist"]["platforms"] = "\n".join(remote.exclude_platforms)


class PythonBanderStage(Stage):
    """
    Python Package Syncing Stage using Bandersnatch
    """

    def __init__(self, remote):
        """Initialize the stage and Bandersnatch config"""
        super().__init__()
        self.remote = remote
        create_bandersnatch_config(remote)

    async def run(self):
        """
        If includes is specified, then only sync those,else try to sync all other packages
        """
        # TODO Change Bandersnatch internal API to take proxy settings in from config parameters
        if self.remote.proxy_url:
            environ['http_proxy'] = self.remote.proxy_url
            environ['https_proxy'] = self.remote.proxy_url
        # Bandersnatch includes leading slash when forming API urls
        url = self.remote.url.rstrip("/")
        # local & global timeouts defaults to 10secs and 5 hours
        async with Master(url) as master:
            deferred_download = self.remote.policy != Remote.IMMEDIATE
            workers = self.remote.download_concurrency or self.remote.DEFAULT_DOWNLOAD_CONCURRENCY
            async with ProgressReport(
                message="Fetching Project Metadata", code="sync.fetching.project"
            ) as p:
                pmirror = PulpMirror(
                    serial=0,  # Serial currently isn't supported by Pulp
                    master=master,
                    workers=workers,
                    deferred_download=deferred_download,
                    python_stage=self,
                    progress_report=p,
                )
                packages_to_sync = None
                if self.remote.includes:
                    packages_to_sync = [
                        Requirement(pkg).name for pkg in self.remote.includes
                    ]
                await pmirror.synchronize(packages_to_sync)


class PulpMirror(Mirror):
    """
    Pulp Mirror Class to perform syncing using Bandersnatch
    """

    def __init__(
        self, serial, master, workers, deferred_download, python_stage, progress_report
    ):
        """Initialize Bandersnatch Mirror"""
        super().__init__(master=master, workers=workers)
        self.synced_serial = serial
        self.python_stage = python_stage
        self.progress_report = progress_report
        self.deferred_download = deferred_download

    async def determine_packages_to_sync(self):
        """
        Calling this means that includes wasn't specified,
        so try to get all of the packages from Mirror (hopefully PyPi)
        """
        number_xmlrpc_attempts = 3
        for attempt in range(number_xmlrpc_attempts):
            logger.info(
                "Attempt {} to get package list from {}".format(
                    attempt, self.master.url
                )
            )
            try:
                if not self.synced_serial:
                    logger.info("Syncing all packages.")
                    # First get the current serial, then start to sync.
                    all_packages = await self.master.all_packages()
                    self.packages_to_sync.update(all_packages)
                    self.target_serial = max(
                        [self.synced_serial] + [int(v) for v in self.packages_to_sync.values()]
                    )
                else:
                    logger.info("Syncing based on changelog.")
                    changed_packages = await self.master.changed_packages(
                        self.synced_serial
                    )
                    self.packages_to_sync.update(changed_packages)
                    self.target_serial = max(
                        [self.synced_serial] + [int(v) for v in self.packages_to_sync.values()]
                    )
                break
            except (ClientError, ClientResponseError, LxmlError):
                # Retry if XMLRPC endpoint failed, server might not support it.
                continue
        else:
            logger.info("Failed to get package list using XMLRPC, trying parse simple page.")
            url = urljoin(self.python_stage.remote.url, "simple/")
            downloader = self.python_stage.remote.get_downloader(url=url)
            result = await downloader.run()
            with open(result.path) as f:
                index = parse_repo_index_page(f.read())
                self.packages_to_sync.update({p: 0 for p in index.projects})
                self.target_serial = result.headers.get(PYPI_LAST_SERIAL, 0)

        self._filter_packages()
        if self.target_serial:
            logger.info(f"Trying to reach serial: {self.target_serial}")
        pkg_count = len(self.packages_to_sync)
        logger.info(f"{pkg_count} packages to sync.")

    async def process_package(self, package):
        """Filters the package and creates content from it"""
        # Don't save anything if our metadata filters all fail.
        await self.progress_report.aincrement()
        if not package.filter_metadata(self.filters.filter_metadata_plugins()):
            return None

        package.filter_all_releases_files(self.filters.filter_release_file_plugins())
        package.filter_all_releases(self.filters.filter_release_plugins())
        await self.create_content(package)

    async def create_content(self, pkg):
        """
        Take the filtered package, separate into releases and
        create a Content Unit to put into the pipeline
        """
        for version, dists in pkg.releases.items():
            for package in dists:
                entry = parse_metadata(pkg.info, version, package)
                url = entry.pop("url")

                artifact = Artifact(sha256=entry["sha256"])
                package = PythonPackageContent(**entry)

                da = DeclarativeArtifact(
                    artifact=artifact,
                    url=url,
                    relative_path=entry["filename"],
                    remote=self.python_stage.remote,
                    deferred_download=self.deferred_download,
                )
                dc = DeclarativeContent(content=package, d_artifacts=[da])

                await self.python_stage.put(dc)

    def finalize_sync(self):
        """No work to be done currently"""
        pass

    def on_error(self, exception, **kwargs):
        """
        TODO
        This should have some error checking
        """
        pass



# coding=utf-8
"""Tests that sync python plugin repositories."""
import unittest

from pulp_smash import config
from pulp_smash.pulp3.bindings import monitor_task, PulpTaskError
from pulp_smash.pulp3.utils import (
    gen_repo,
    get_added_content_summary,
    get_removed_content_summary,
    get_content_summary,
)

from pulp_python.tests.functional.constants import (
    PYTHON_XS_FIXTURE_SUMMARY,
    PYTHON_XS_PACKAGE_COUNT,
    PYTHON_INVALID_FIXTURE_URL,
    PYTHON_WITHOUT_PRERELEASE_FIXTURE_SUMMARY,
    PYTHON_PRERELEASE_TEST_SPECIFIER,
    PYTHON_WITH_PRERELEASE_FIXTURE_SUMMARY,
    PYTHON_WITH_PRERELEASE_COUNT,
    PYTHON_WITHOUT_PRERELEASE_COUNT,
    PYTHON_CONTENT_NAME,
    PYTHON_XS_PROJECT_SPECIFIER,
    PYTHON_MD_PROJECT_SPECIFIER,
    PYTHON_MD_FIXTURE_SUMMARY,
    PYTHON_MD_PACKAGE_COUNT,
    PYTHON_SM_PROJECT_SPECIFIER,
    PYTHON_SM_PACKAGE_COUNT,
    PYTHON_UNAVAILABLE_PACKAGE_COUNT,
    PYTHON_UNAVAILABLE_PROJECT_SPECIFIER,
    PYTHON_LG_PROJECT_SPECIFIER,
    PYTHON_LG_FIXTURE_SUMMARY,
    PYTHON_LG_FIXTURE_COUNTS,
    DJANGO_LATEST_3,
    SCIPY_COUNTS,
)
from pulp_python.tests.functional.utils import gen_python_client, gen_python_remote
from pulp_python.tests.functional.utils import set_up_module as setUpModule  # noqa:F401

from pulpcore.client.pulp_python import (
    RepositoriesPythonApi,
    RepositorySyncURL,
    RemotesPythonApi,
)


class BasicSyncTestCase(unittest.TestCase):
    """Sync a repository with the python plugin."""

    @classmethod
    def setUpClass(cls):
        """Create class-wide variables."""
        cls.cfg = config.get_config()
        cls.client = gen_python_client()

    def test_sync(self):
        """Sync repositories with the python plugin.

        In order to sync a repository a remote has to be associated within
        this repository. When a repository is created this version field is set
        as None. After a sync the repository version is updated.

        Do the following:

        1. Create a repository, and a remote.
        2. Assert that repository version is None.
        3. Sync the remote.
        4. Assert that repository version is not None.
        5. Assert that the correct number of units were added and are present
           in the repo.
        6. Sync the remote one more time.
        7. Assert that repository version is different from the previous one.
        8. Assert that the same number of are present and that no units were
           added.
        """
        repo_api = RepositoriesPythonApi(self.client)
        remote_api = RemotesPythonApi(self.client)

        repo = repo_api.create(gen_repo())
        self.addCleanup(repo_api.delete, repo.pulp_href)

        body = gen_python_remote()
        remote = remote_api.create(body)
        self.addCleanup(remote_api.delete, remote.pulp_href)

        # Sync the repository.
        self.assertEqual(repo.latest_version_href, f"{repo.pulp_href}versions/0/")
        repository_sync_data = RepositorySyncURL(remote=remote.pulp_href)
        sync_response = repo_api.sync(repo.pulp_href, repository_sync_data)
        monitor_task(sync_response.task)
        repo = repo_api.read(repo.pulp_href)

        self.assertIsNotNone(repo.latest_version_href)
        self.assertDictEqual(
            get_content_summary(repo.to_dict()), PYTHON_XS_FIXTURE_SUMMARY
        )
        self.assertDictEqual(
            get_added_content_summary(repo.to_dict()), PYTHON_XS_FIXTURE_SUMMARY
        )

        # Sync the repository again.
        latest_version_href = repo.latest_version_href
        repository_sync_data = RepositorySyncURL(remote=remote.pulp_href)
        sync_response = repo_api.sync(repo.pulp_href, repository_sync_data)
        monitor_task(sync_response.task)
        repo = repo_api.read(repo.pulp_href)

        self.assertEqual(latest_version_href, repo.latest_version_href)
        self.assertDictEqual(
            get_content_summary(repo.to_dict()), PYTHON_XS_FIXTURE_SUMMARY
        )


class SyncInvalidTestCase(unittest.TestCase):
    """Sync a repository with a given url on the remote."""

    @classmethod
    def setUpClass(cls):
        """Create class-wide variables."""
        cls.client = gen_python_client()

    @unittest.skip("Sync task does no work with invalid url")
    def test_invalid_url(self):
        """Sync a repository using a remote url that does not exist.

        Test that we get a task failure. See :meth:`do_test`.
        """
        with self.assertRaises(PulpTaskError) as cm:
            self.do_test("http://i-am-an-invalid-url.com/invalid/")
        task = cm.exception.task.to_dict()
        self.assertIsNotNone(task["error"]["description"])

    # Provide an invalid repository and specify keywords in the anticipated error message
    @unittest.skip(
        "Need a fixture url with invalid content"
    )  # This test is not present in the original test_sync
    def test_invalid_python_content(self):
        """Sync a repository using an invalid plugin_content repository.

        Assert that an exception is raised, and that error message has
        keywords related to the reason of the failure. See :meth:`do_test`.
        """
        with self.assertRaises(PulpTaskError) as cm:
            self.do_test(PYTHON_INVALID_FIXTURE_URL)
        task = cm.exception.task.to_dict()
        # this url is not valid on the fixture,
        # it needs to point to somewhere that has invalid content
        for key in ("mismached", "empty"):
            self.assertIn(key, task["error"]["description"])

    def do_test(self, url):
        """Sync a repository given ``url`` on the remote."""
        repo_api = RepositoriesPythonApi(self.client)
        remote_api = RemotesPythonApi(self.client)

        repo = repo_api.create(gen_repo())
        self.addCleanup(repo_api.delete, repo.pulp_href)

        body = gen_python_remote(url=url)
        remote = remote_api.create(body)
        self.addCleanup(remote_api.delete, remote.pulp_href)

        repository_sync_data = RepositorySyncURL(remote=remote.pulp_href)
        sync_response = repo_api.sync(repo.pulp_href, repository_sync_data)
        return monitor_task(sync_response.task)


class PrereleasesTestCase(unittest.TestCase):
    """
    Sync a repository with and without prereleases included by the Remote.
    """

    @classmethod
    def setUpClass(cls):
        """
        Create class-wide variables.
        """
        cls.client = gen_python_client()
        cls.remote = None
        cls.remote_api = RemotesPythonApi(cls.client)
        cls.repo_api = RepositoriesPythonApi(cls.client)
        cls.repo = cls.repo_api.create(gen_repo())

    @classmethod
    def tearDownClass(cls):
        """
        Clean class-wide variable.
        """
        cls.repo_api.delete(cls.repo.pulp_href)
        cls.remote_api.delete(cls.remote.pulp_href)

    def test_01_excluding_prereleases(self):
        """
        Sync a Remote, excluding prereleases.

        Do the following:

        1. Create a remote with prereleases=False.
        2. Sync the remote.
        3. Assert that the content counts in the repo match the non-prerelease packages matched
           by the specifiers.

        """
        body = gen_python_remote(
            includes=PYTHON_PRERELEASE_TEST_SPECIFIER, prereleases=False
        )
        sync_to_remote(self, body, create=True)

        self.assertDictEqual(
            get_content_summary(self.repo.to_dict()),
            PYTHON_WITHOUT_PRERELEASE_FIXTURE_SUMMARY,
        )

    def test_02_including_prereleases(self):
        """
        Sync a Remote, including prereleases.

        Do the following:

        1. Update the remote to include pre-releases.
        2. Sync the remote again.
        3. Assert that the content counts in the repo match *all* the packages matched by the
           specifier, including prereleases.

        """
        body = {"prereleases": True}
        sync_to_remote(self, body)

        self.assertEqual(
            get_content_summary(self.repo.to_dict()),
            PYTHON_WITH_PRERELEASE_FIXTURE_SUMMARY,
        )
        self.assertEqual(
            get_added_content_summary(self.repo.to_dict())[PYTHON_CONTENT_NAME],
            PYTHON_WITH_PRERELEASE_COUNT - PYTHON_WITHOUT_PRERELEASE_COUNT,
        )

    def test_03_removing_units(self):
        """
        Sync a Remote, excluding prereleases again.

        Just to be sure that the units are being properly removed afterwards.

        Do the following:

        1. Update the remote to exclude pre-releases again.
        2. Sync the remote again.
        3. Assert that we're back to the state in test_01_excluding_prereleases.

        """
        body = {"prereleases": False}
        sync_to_remote(self, body, mirror=True)

        self.assertDictEqual(
            get_content_summary(self.repo.to_dict()),
            PYTHON_WITHOUT_PRERELEASE_FIXTURE_SUMMARY,
        )
        self.assertEqual(
            get_removed_content_summary(self.repo.to_dict())[PYTHON_CONTENT_NAME],
            PYTHON_WITH_PRERELEASE_COUNT - PYTHON_WITHOUT_PRERELEASE_COUNT,
        )


class IncludesExcludesTestCase(unittest.TestCase):
    """
    Test behavior of the includes and excludes fields on the Remote during sync.
    """

    @classmethod
    def setUpClass(cls):
        """Create class-wide variables."""
        cls.client = gen_python_client()

        cls.remote = {}
        cls.remote_api = RemotesPythonApi(cls.client)
        cls.repo_api = RepositoriesPythonApi(cls.client)
        cls.repo = cls.repo_api.create(gen_repo())

    @classmethod
    def tearDownClass(cls):
        """Clean class-wide variable."""
        cls.repo_api.delete(cls.repo.pulp_href)
        cls.remote_api.delete(cls.remote.pulp_href)

    def test_01_basic_include(self):
        """
        Test updating the remote and performing normal sync.
        Really just a setup for the next test.
        Do the following:
        1. Create a remote.
        2. Sync the remote.
        3. Assert that the content counts in the repo match the correct count for the specifier.
        """
        body = gen_python_remote(includes=PYTHON_XS_PROJECT_SPECIFIER)
        sync_to_remote(self, body, create=True)

        self.assertDictEqual(
            get_content_summary(self.repo.to_dict()), PYTHON_XS_FIXTURE_SUMMARY
        )

    def test_02_add_superset_include(self):
        """
        Test updating the remote with a larger set of includes, and syncing again.
        Do the following:
        1. Update the remote with a larger specifier that is a strict superset of the previous one.
        2. Sync the remote again.
        3. Assert that the content counts in the repo increased to match the new correct count.
        """
        body = {"includes": PYTHON_MD_PROJECT_SPECIFIER}
        sync_to_remote(self, body)

        self.assertDictEqual(
            get_content_summary(self.repo.to_dict()), PYTHON_MD_FIXTURE_SUMMARY
        )

    def test_03_add_subset_exclude(self):
        """
        Test that excluding a subset of the packages will reduce the count by that amount.
        Do the following:
        1. Update the remote to exclude a specifier that is a strict subset of the previous one.
        2. Sync the remote again.
        3. Assert that that the content counts in the repo decreased by the count of the
           excludes specifier.
        """
        body = {"excludes": PYTHON_SM_PROJECT_SPECIFIER}
        sync_to_remote(self, body, mirror=True)

        self.assertEqual(
            get_content_summary(self.repo.to_dict())[PYTHON_CONTENT_NAME],
            PYTHON_MD_PACKAGE_COUNT - PYTHON_SM_PACKAGE_COUNT,
        )
        self.assertEqual(
            get_removed_content_summary(self.repo.to_dict())[PYTHON_CONTENT_NAME],
            PYTHON_SM_PACKAGE_COUNT,
        )

    def test_04_remove_excludes(self):
        """
        Test that removing some of the excludes increases the count again.
        Do the following:
        1. Update the remote to exclude a specifier that is smaller than the previous one.
        2. Sync the remote again.
        3. Assert that the content counts in the repo have increased again, to the count
           of the smaller excludes specifier.
        """
        body = {"excludes": PYTHON_XS_PROJECT_SPECIFIER}
        sync_to_remote(self, body, mirror=True)

        self.assertEqual(
            get_content_summary(self.repo.to_dict())[PYTHON_CONTENT_NAME],
            PYTHON_MD_PACKAGE_COUNT - PYTHON_XS_PACKAGE_COUNT,
        )


class UnavailableProjectsTestCase(unittest.TestCase):
    """
    Test syncing with projects that aren't on the upstream remote.

    Tests that sync doesn't fail if the Remote contains projects (includes or excludes) for which
    metadata does not exist on the upstream remote.
    """

    @classmethod
    def setUpClass(cls):
        """Create class-wide variables."""
        cls.client = gen_python_client()
        cls.repo_api = RepositoriesPythonApi(cls.client)
        cls.remote_api = RemotesPythonApi(cls.client)

    @classmethod
    def setUp(cls):
        """Create class-wide variables per test"""
        cls.repo = cls.repo_api.create(gen_repo())
        cls.remote = None

    @classmethod
    def tearDown(cls):
        """Destroy class-wide variables per test"""
        cls.remote_api.delete(cls.remote.pulp_href)
        cls.repo_api.delete(cls.repo.pulp_href)

    def test_include_unavailable_projects(self):
        """
        Test that the sync doesn't fail if some included projects aren't available.
        Do the following:
        1. Create a remote.
        2. Sync the remote.
        3. Assert that the content counts in the repo match the correct count for the specifier.
        """
        body = gen_python_remote(includes=PYTHON_UNAVAILABLE_PROJECT_SPECIFIER)
        sync_to_remote(self, body, create=True)

        self.assertEqual(
            get_content_summary(self.repo.to_dict())[PYTHON_CONTENT_NAME],
            PYTHON_UNAVAILABLE_PACKAGE_COUNT,
        )

    def test_exclude_unavailable_projects(self):
        """
        Test that sync doesn't fail if some of the excluded projects aren't available.
        Do the following:
        1. Update the remote to exclude a specifier that is smaller than the previous one.
        2. Sync the remote again.
        3. Assert that the content counts in the repo have increased again, to the count
           of the smaller excludes specifier.
        """
        body = gen_python_remote(
            includes=PYTHON_MD_PROJECT_SPECIFIER,
            excludes=PYTHON_UNAVAILABLE_PROJECT_SPECIFIER,
        )
        sync_to_remote(self, body, create=True)

        self.assertEqual(
            get_content_summary(self.repo.to_dict())[PYTHON_CONTENT_NAME],
            PYTHON_MD_PACKAGE_COUNT - PYTHON_UNAVAILABLE_PACKAGE_COUNT,
        )


class LatestKeptPackagesTestCase(unittest.TestCase):
    """
    Test checks that latest X packages are synced when latest kept is
    specified

    Targets issue:
    https://pulp.plan.io/issues/138

    This feature uses Bandersnatch's latest_kept filter which doesn't work well
    with the other filters like pre-releases and allow/blocklist. Whether to
    count the behavior as a bug or feature is hard to tell.
    """

    @classmethod
    def setUpClass(cls):
        """Create class-wide variables."""
        cls.client = gen_python_client()
        cls.repo_api = RepositoriesPythonApi(cls.client)
        cls.remote_api = RemotesPythonApi(cls.client)

    @classmethod
    def setUp(cls):
        """Create class-wide variables per test"""
        cls.repo = cls.repo_api.create(gen_repo())
        cls.remote = None

    @classmethod
    def tearDown(cls):
        """Destroy class-wide variables per test"""
        cls.remote_api.delete(cls.remote.pulp_href)
        cls.repo_api.delete(cls.repo.pulp_href)

    def test_latest_kept_sync(self):
        """
        Tests latest_kept on syncing one package w/ prereleases
        """
        body = gen_python_remote(
            includes=["Django"],
            keep_latest_packages=3,
            prereleases=True,
        )
        sync_to_remote(self, body, create=True)

        self.assertEqual(
            get_content_summary(self.repo.to_dict())[PYTHON_CONTENT_NAME],
            DJANGO_LATEST_3,
        )

    def test_latest_kept_sync_all(self):
        """
        Tests latest_kept on syncing multiple packages w/ prereleases
        """
        body = gen_python_remote(
            includes=PYTHON_LG_PROJECT_SPECIFIER,
            keep_latest_packages=3,
            prereleases=True,
        )
        sync_to_remote(self, body, create=True)

        self.assertEqual(
            get_content_summary(self.repo.to_dict())[PYTHON_CONTENT_NAME],
            PYTHON_LG_FIXTURE_COUNTS["latest_3"]
        )


class PackageTypeTestCase(unittest.TestCase):
    """
    Tests that check only specified package types can be synced

    Targets:
    https://pulp.plan.io/issues/2040
    """

    @classmethod
    def setUpClass(cls):
        """Create class-wide variables."""
        cls.client = gen_python_client()
        cls.repo_api = RepositoriesPythonApi(cls.client)
        cls.remote_api = RemotesPythonApi(cls.client)

    @classmethod
    def setUp(cls):
        """Create class-wide variables per test"""
        cls.repo = cls.repo_api.create(gen_repo())
        cls.remote = None

    @classmethod
    def tearDown(cls):
        """Destroy class-wide variables per test"""
        cls.remote_api.delete(cls.remote.pulp_href)
        cls.repo_api.delete(cls.repo.pulp_href)

    def test_sdist_sync_only(self):
        """Checks that only sdist content is synced"""
        body = gen_python_remote(
            includes=PYTHON_LG_PROJECT_SPECIFIER,
            package_types=["sdist"],
            prereleases=True,
        )
        sync_to_remote(self, body, create=True)

        self.assertEqual(
            get_content_summary(self.repo.to_dict())[PYTHON_CONTENT_NAME],
            PYTHON_LG_FIXTURE_COUNTS["sdist"]
        )

    def test_bdist_wheel_sync_only(self):
        """Checks that only bdist_wheel content is synced"""
        body = gen_python_remote(
            includes=PYTHON_LG_PROJECT_SPECIFIER,
            package_types=["bdist_wheel"],
            prereleases=True,
        )
        sync_to_remote(self, body, create=True)

        self.assertEqual(
            get_content_summary(self.repo.to_dict())[PYTHON_CONTENT_NAME],
            PYTHON_LG_FIXTURE_COUNTS["bdist_wheel"]
        )

    def test_both_together_sync(self):
        """Checks that specifying sdist and bdist_wheel gets all packages"""
        body = gen_python_remote(
            includes=PYTHON_LG_PROJECT_SPECIFIER,
            package_types=["bdist_wheel", "sdist"],
            prereleases=True,
        )
        sync_to_remote(self, body, create=True)

        self.assertEqual(
            get_content_summary(self.repo.to_dict()),
            PYTHON_LG_FIXTURE_SUMMARY
        )


class PlatformExcludeTestCase(unittest.TestCase):
    """
    Tests for platform specific packages not being synced when specified

    Only our scipy packages have platform specific versions
    23 release files:
    4 macos releases
    8 windows releases
    10 linux releases
    1 any platform release
    """

    @classmethod
    def setUpClass(cls):
        """Create class-wide variables."""
        cls.client = gen_python_client()
        cls.repo_api = RepositoriesPythonApi(cls.client)
        cls.remote_api = RemotesPythonApi(cls.client)

    @classmethod
    def setUp(cls):
        """Create class-wide variables per test"""
        cls.repo = cls.repo_api.create(gen_repo())
        cls.remote = None

    @classmethod
    def tearDown(cls):
        """Destroy class-wide variables per test"""
        cls.remote_api.delete(cls.remote.pulp_href)
        cls.repo_api.delete(cls.repo.pulp_href)

    def test_no_windows_sync(self):
        """Tests that no windows packages are synced"""
        body = gen_python_remote(
            includes=["scipy"],
            exclude_platforms=["windows"],
            prereleases=True,
        )
        sync_to_remote(self, body, create=True)

        self.assertEqual(
            get_content_summary(self.repo.to_dict())[PYTHON_CONTENT_NAME],
            SCIPY_COUNTS["total"] - SCIPY_COUNTS["windows"]
        )

    def test_no_macos_sync(self):
        """Tests that no macos packages are synced"""
        body = gen_python_remote(
            includes=["scipy"],
            exclude_platforms=["macos"],
            prereleases=True,
        )
        sync_to_remote(self, body, create=True)

        self.assertEqual(
            get_content_summary(self.repo.to_dict())[PYTHON_CONTENT_NAME],
            SCIPY_COUNTS["total"] - SCIPY_COUNTS["macos"]
        )

    def test_no_linux_sync(self):
        """Tests that no linux packages are synced"""
        body = gen_python_remote(
            includes=["scipy"],
            exclude_platforms=["linux"],
            prereleases=True,
        )
        sync_to_remote(self, body, create=True)

        self.assertEqual(
            get_content_summary(self.repo.to_dict())[PYTHON_CONTENT_NAME],
            SCIPY_COUNTS["total"] - SCIPY_COUNTS["linux"]
        )

    def test_no_platform_sync(self):
        """Tests that no package specified for a platform is synced"""
        body = gen_python_remote(
            includes=["scipy"],
            exclude_platforms=["linux", "windows", "macos", "freebsd"],
            prereleases=True,
        )
        sync_to_remote(self, body, create=True)

        self.assertEqual(
            get_content_summary(self.repo.to_dict())[PYTHON_CONTENT_NAME],
            SCIPY_COUNTS["no_os"]
        )


def sync_to_remote(self, body, create=False, mirror=False):
    """Takes a body and creates/updates a remote object, then it performs a sync"""
    if create:
        type(self).remote = self.remote_api.create(body)
    else:
        remote_task = self.remote_api.partial_update(self.remote.pulp_href, body)
        monitor_task(remote_task.task)
        type(self).remote = self.remote_api.read(self.remote.pulp_href)

    repository_sync_data = RepositorySyncURL(
        remote=self.remote.pulp_href, mirror=mirror
    )
    sync_response = self.repo_api.sync(self.repo.pulp_href, repository_sync_data)
    monitor_task(sync_response.task)
    type(self).repo = self.repo_api.read(self.repo.pulp_href)



# coding=utf-8
"""Tests that CRUD python remotes."""
from random import choice
import unittest

from pulp_smash import utils

from pulp_smash.pulp3.bindings import monitor_task
from pulp_smash.pulp3.constants import ON_DEMAND_DOWNLOAD_POLICIES

from pulp_python.tests.functional.utils import (
    gen_python_client,
    gen_python_remote,
    skip_if,
)
from pulp_python.tests.functional.utils import set_up_module as setUpModule  # noqa:F401

from pulpcore.client.pulp_python import RemotesPythonApi
from pulpcore.client.pulp_python.exceptions import ApiException

from pulp_python.tests.functional.constants import (
    BANDERSNATCH_CONF,
    DEFAULT_BANDER_REMOTE_BODY,
    PYTHON_INVALID_SPECIFIER_NO_NAME,
    PYTHON_INVALID_SPECIFIER_BAD_VERSION,
    PYTHON_VALID_SPECIFIER_NO_VERSION,
)
from tempfile import NamedTemporaryFile


class CRUDRemotesTestCase(unittest.TestCase):
    """CRUD remotes."""

    @classmethod
    def setUpClass(cls):
        """Create class-wide variables."""
        cls.remote_api = RemotesPythonApi(gen_python_client())

    def test_01_create_remote(self):
        """Create a remote."""
        body = _gen_verbose_remote()
        type(self).remote = self.remote_api.create(body)
        for key in ("username", "password"):
            del body[key]
        for key, val in body.items():
            with self.subTest(key=key):
                self.assertEqual(self.remote.to_dict()[key], val, key)

    @skip_if(bool, "remote", False)
    def test_02_create_same_name(self):
        """Try to create a second remote with an identical name.

        See: `Pulp Smash #1055
        <https://github.com/pulp/pulp-smash/issues/1055>`_.
        """
        body = gen_python_remote()
        body["name"] = self.remote.name
        with self.assertRaises(ApiException):
            self.remote_api.create(body)

    @skip_if(bool, "remote", False)
    def test_02_read_remote(self):
        """Read a remote by its href."""
        remote = self.remote_api.read(self.remote.pulp_href)
        for key, val in self.remote.to_dict().items():
            with self.subTest(key=key):
                self.assertEqual(remote.to_dict()[key], val, key)

    @skip_if(bool, "remote", False)
    def test_02_read_remotes(self):
        """Read a remote by its name."""
        page = self.remote_api.list(name=self.remote.name)
        self.assertEqual(len(page.results), 1)
        for key, val in self.remote.to_dict().items():
            with self.subTest(key=key):
                self.assertEqual(page.results[0].to_dict()[key], val, key)

    @skip_if(bool, "remote", False)
    def test_03_partially_update(self):
        """Update a remote using HTTP PATCH."""
        body = _gen_verbose_remote()
        response = self.remote_api.partial_update(self.remote.pulp_href, body)
        monitor_task(response.task)
        for key in ("username", "password"):
            del body[key]
        type(self).remote = self.remote_api.read(self.remote.pulp_href)
        for key, val in body.items():
            with self.subTest(key=key):
                self.assertEqual(self.remote.to_dict()[key], val, key)

    @skip_if(bool, "remote", False)
    def test_04_fully_update(self):
        """Update a remote using HTTP PUT."""
        body = _gen_verbose_remote()
        response = self.remote_api.update(self.remote.pulp_href, body)
        monitor_task(response.task)
        for key in ("username", "password"):
            del body[key]
        type(self).remote = self.remote_api.read(self.remote.pulp_href)
        for key, val in body.items():
            with self.subTest(key=key):
                self.assertEqual(self.remote.to_dict()[key], val, key)

    @skip_if(bool, "remote", False)
    def test_05_delete(self):
        """Delete a remote."""
        response = self.remote_api.delete(self.remote.pulp_href)
        monitor_task(response.task)
        with self.assertRaises(ApiException):
            self.remote_api.read(self.remote.pulp_href)


class CreateRemoteFromBandersnatchConfig(unittest.TestCase):
    """
    Verify whether it's possible to create a remote from a Bandersnatch config

    This test targets the following issues:

    * `Pulp #6929 <https://pulp.plan.io/issues/6929>`_
    * `Pulp #7331 <https://pulp.plan.io/issues/7331>`_
    """

    def test_01_creation(self):
        """Create a remote from Bandersnatch config."""
        remote_api = RemotesPythonApi(gen_python_client())
        with NamedTemporaryFile() as config:
            config.write(BANDERSNATCH_CONF)
            config.seek(0)
            name = utils.uuid4()
            remote = remote_api.from_bandersnatch(config.name, name)
            self.addCleanup(remote_api.delete, remote.pulp_href)
            expected = _gen_expected_remote_body(name)
            for key, val in expected.items():
                with self.subTest(key=key):
                    self.assertEqual(remote.to_dict()[key], val, key)

    def test_02_ondemand(self):
        """Create a remote from Bandersnatch config w/ policy remote"""
        remote_api = RemotesPythonApi(gen_python_client())
        with NamedTemporaryFile() as config:
            config.write(BANDERSNATCH_CONF)
            config.seek(0)
            name = utils.uuid4()
            policy = "on_demand"
            remote = remote_api.from_bandersnatch(config.name, name, policy=policy)
            self.addCleanup(remote_api.delete, remote.pulp_href)
            expected = _gen_expected_remote_body(name, policy=policy)
            for key, val in expected.items():
                with self.subTest(key=key):
                    self.assertEqual(remote.to_dict()[key], val, key)


class CreateRemoteNoURLTestCase(unittest.TestCase):
    """Verify whether is possible to create a remote without a URL."""

    def test_all(self):
        """Verify whether is possible to create a remote without a URL.

        This test targets the following issues:

        * `Pulp #3395 <https://pulp.plan.io/issues/3395>`_
        * `Pulp Smash #984 <https://github.com/pulp/pulp-smash/issues/984>`_
        """
        body = gen_python_remote()
        del body["url"]
        with self.assertRaises(ApiException):
            RemotesPythonApi(gen_python_client()).create(body)


class RemoteDownloadPolicyTestCase(unittest.TestCase):
    """Verify download policy behavior for valid and invalid values.

    In Pulp 3, there are are different download policies.

    This test targets the following testing scenarios:

    1. Creating a remote without a download policy.
       Verify the creation is successful and immediate it is policy applied.
    2. Change the remote policy from default.
       Verify the change is successful.
    3. Attempt to change the remote policy to an invalid string.
       Verify an ApiException is given for the invalid policy as well
       as the policy remaining unchanged.

    For more information on the remote policies, see the Pulp3
    API on an installed server:

    * /pulp/api/v3/docs/#operation`

    This test targets the following issues:

    * `Pulp #4420 <https://pulp.plan.io/issues/4420>`_
    * `Pulp #3763 <https://pulp.plan.io/issues/3763>`_
    """

    @classmethod
    def setUpClass(cls):
        """Create class-wide variables."""
        cls.remote_api = RemotesPythonApi(gen_python_client())
        cls.remote = {}
        cls.policies = ON_DEMAND_DOWNLOAD_POLICIES
        cls.body = _gen_verbose_remote()

    @classmethod
    def tearDownClass(cls):
        """Clean class-wide variable."""
        results = cls.remote_api.list().to_dict()["results"]
        for result in results:
            cls.remote_api.delete(result["pulp_href"])

    def test_01_no_defined_policy(self):
        """Verify the default policy `on_demand`.

        When no policy is defined, the default policy of `on_demand`
        is applied.
        """
        del self.body["policy"]
        self.remote.update(self.remote_api.create(self.body).to_dict())
        self.assertEqual(self.remote["policy"], "on_demand", self.remote)

    @skip_if(len, "policies", 1)
    def test_02_change_policy(self):
        """Verify ability to change policy to value other than the default.

        Update the remote policy to a valid value other than `on_demand`
        and verify the new set value.
        """
        changed_policy = choice([item for item in self.policies if item != "on_demand"])
        response = self.remote_api.partial_update(
            self.remote["pulp_href"], {"policy": changed_policy}
        )
        monitor_task(response.task)
        self.remote.update(self.remote_api.read(self.remote["pulp_href"]).to_dict())
        self.assertEqual(self.remote["policy"], changed_policy, self.remote)

    @skip_if(bool, "remote", False)
    def test_03_invalid_policy(self):
        """Verify an invalid policy does not update the remote policy.

        Get the current remote policy.
        Attempt to update the remote policy to an invalid value.
        Verify the policy remains the same.
        """
        remote = self.remote_api.read(self.remote["pulp_href"]).to_dict()
        with self.assertRaises(ApiException):
            self.remote_api.partial_update(
                self.remote["pulp_href"], {"policy": utils.uuid4()}
            )
        self.remote.update(self.remote_api.read(self.remote["pulp_href"]).to_dict())
        self.assertEqual(remote["policy"], self.remote["policy"], self.remote)


class CreateRemoteWithInvalidProjectSpecifiersTestCase(unittest.TestCase):
    """
    Test that creating a remote with an invalid project specifier fails.
    """

    @classmethod
    def setUpClass(cls):
        """
        Create class-wide variables.
        """
        cls.remote_api = RemotesPythonApi(gen_python_client())

    def test_includes_with_no_name(self):
        """
        Test an include specifier without a "name" field.
        """
        body = gen_python_remote(includes=PYTHON_INVALID_SPECIFIER_NO_NAME)
        with self.assertRaises(ApiException):
            self.remote_api.create(body)

    def test_includes_with_bad_version(self):
        """
        Test an include specifier with an invalid "version_specifier" field value.
        """
        body = gen_python_remote(includes=PYTHON_INVALID_SPECIFIER_BAD_VERSION)
        with self.assertRaises(ApiException):
            self.remote_api.create(body)

    def test_excludes_with_no_name(self):
        """
        Test an exclude specifier without a "name" field.
        """
        body = gen_python_remote(excludes=PYTHON_INVALID_SPECIFIER_NO_NAME)
        with self.assertRaises(ApiException):
            self.remote_api.create(body)

    def test_excludes_with_bad_version(self):
        """
        Test an exclude specifier with an invalid "version_specifier" field value.
        """
        body = gen_python_remote(excludes=PYTHON_INVALID_SPECIFIER_BAD_VERSION)
        with self.assertRaises(ApiException):
            self.remote_api.create(body)


class CreateRemoteWithNoVersionTestCase(unittest.TestCase):
    """
    Test that creating a remote with no "version_specifier" on the project specifier works.
    """

    @classmethod
    def setUpClass(cls):
        """
        Create class-wide variables.
        """
        cls.remote_api = RemotesPythonApi(gen_python_client())

    def test_includes_with_no_version(self):
        """
        Test an include specifier without a "version_specifier" field.
        """
        body = gen_python_remote(includes=PYTHON_VALID_SPECIFIER_NO_VERSION)
        remote = self.remote_api.create(body).to_dict()
        self.addCleanup(self.remote_api.delete, remote["pulp_href"])

        self.assertEqual(remote["includes"][0], PYTHON_VALID_SPECIFIER_NO_VERSION[0])

    def test_excludes_with_no_version(self):
        """
        Test an exclude specifier without a "version_specifier" field.
        """
        body = gen_python_remote(excludes=PYTHON_VALID_SPECIFIER_NO_VERSION)
        remote = self.remote_api.create(body).to_dict()
        self.addCleanup(self.remote_api.delete, remote["pulp_href"])

        self.assertEqual(remote["includes"][0], PYTHON_VALID_SPECIFIER_NO_VERSION[0])
        self.assertEqual(remote["excludes"][0], PYTHON_VALID_SPECIFIER_NO_VERSION[0])


class UpdateRemoteWithInvalidProjectSpecifiersTestCase(unittest.TestCase):
    """
    Test that updating a remote with an invalid project specifier fails non-destructively.
    """

    @classmethod
    def setUpClass(cls):
        """
        Create class-wide variables.
        """
        cls.remote_api = RemotesPythonApi(gen_python_client())
        cls.remote = cls.remote_api.create(gen_python_remote())
        cls._original_remote = cls.remote

    @classmethod
    def tearDownClass(cls):
        """
        Clean class-wide variable.
        """
        cls.remote_api.delete(cls.remote.pulp_href)

    def test_includes_with_no_name(self):
        """
        Test an include specifier without a "name" field.
        """
        body = {"includes": PYTHON_INVALID_SPECIFIER_NO_NAME}
        with self.assertRaises(ApiException):
            self.remote_api.partial_update(self.remote.pulp_href, body)

    def test_includes_with_bad_version(self):
        """
        Test an include specifier with an invalid "version_specifier" field value.
        """
        body = {"includes": PYTHON_INVALID_SPECIFIER_BAD_VERSION}
        with self.assertRaises(ApiException):
            self.remote_api.partial_update(self.remote.pulp_href, body)

    def test_excludes_with_no_name(self):
        """
        Test an exclude specifier without a "name" field.
        """
        body = {"excludes": PYTHON_INVALID_SPECIFIER_NO_NAME}
        with self.assertRaises(ApiException):
            self.remote_api.partial_update(self.remote.pulp_href, body)

    def test_excludes_with_bad_version(self):
        """
        Test an exclude specifier with an invalid "version_specifier" field value.
        """
        body = {"excludes": PYTHON_INVALID_SPECIFIER_BAD_VERSION}
        with self.assertRaises(ApiException):
            self.remote_api.partial_update(self.remote.pulp_href, body)


def _gen_verbose_remote():
    """Return a semi-random dict for use in defining a remote.

    For most tests, it"s desirable to create remotes with as few attributes
    as possible, so that the tests can specifically target and attempt to break
    specific features. This module specifically targets remotes, so it makes
    sense to provide as many attributes as possible.

    Note that 'username' and 'password' are write-only attributes.
    """
    attrs = gen_python_remote()
    attrs.update(
        {
            "password": utils.uuid4(),
            "username": utils.uuid4(),
            "policy": choice(ON_DEMAND_DOWNLOAD_POLICIES),
        }
    )
    return attrs


def _gen_expected_remote_body(name, **kwargs):
    """Generates a remote body based on names and dictionary in kwargs"""
    # The defaults found in bandersnatch_conf
    body = DEFAULT_BANDER_REMOTE_BODY
    body["name"] = name
    # overwrite the defaults if specified in kwargs
    body.update(kwargs)
    return body
